/*tree29.c
   maeyeda's 2-tree method to evaluate topological formulas using
   (minty's algorithm for spanning tree enumeration
   in symbolic network analysis, lin) */

#define gcc 0
#if gcc
#include <stdio.h>
#else
#include "..\include\stdio.h"
#endif

#define dynamic 1

#define actmax 20		/* max active 2-ports */
#define pmax 60                 /*max passive (one-port) edges*/
#define emax (pmax+actmax)	/* max edges */
#define nmax 30			/* max nodes */
#define stk1 nmax		/* minty alg. stack */
#define stk2 nmax		/* traverse stack */
#if dynamic
#else
#define tbuf1max 100000		/*v-tree  list to match i-trees */
#define tbuf2max 100000
#endif
#define vtree 1
#define itree 2

/* numerator and denominator indicators */
#define V     0
#define wijkl 1
#define wijij 2
#define wklkl 3
#define wklij 4
#define uijkl 5

/* test parameter masks */
#define INPUT_DUMP 1
#define RAW_TREES 2
#define BUFF_TREES 4
#define TREE_PERM 8
#define MIXED_UNITS 16
#define ADMITT_UNITS 32
#define PRINT_STATS 64

#if dynamic
int tbuf1max, tbuf2max;
int (*treebuf)[nmax];
int (*treebuf2)[nmax];
int array[nmax];
#else
int treebuf[tbuf1max][nmax];
int treebuf2[tbuf2max][nmax];
#endif
int tbuf1overage,tbuf2overage;
int nbuf;
int branchtag[emax];
int nbuf2, dtag;
int ni,nj,nk,nl,ni1,nj1,nk1,nl1,opampcnt, n, int_node, nodelist[nmax];
int p0[pmax], p1[pmax], p2[pmax];
int test;

int e, p, r, a, d0[emax], d1[emax], d2[emax], dv1[emax], dv2[emax];
int a0[actmax], a1[actmax], a2[actmax],a3[actmax], a4[actmax], branches;
char branch[emax][8];



#define maxsym emax
#define maxterm 20
#define maxdegree nmax
#if dynamic
int (*numtermlist)[maxsym+1];
int (*denomtermlist)[maxsym+1];
int (*numpoly)[maxdegree],(*denompoly)[maxdegree];
#else
int numtermlist[maxterm * maxdegree][maxsym+1];
int denomtermlist[maxterm * maxdegree][maxsym+1];
int numpoly[maxdegree][maxterm],denompoly[maxdegree][maxterm];
#endif
int term[maxsym],cancelfactor[maxsym],cancelcoeff,cancelsign,cancelexp;
int numtoterms;
int denomtoterms;
int termlistoverage;
int numdegterms[maxdegree];
int denomdegterms[maxdegree];
int nummaxterms,nummaxpoly,denommaxterms,denommaxpoly;
#define NUM_VARS &numtermlist,&numpoly,numdegterms,&numtoterms,&nummaxterms,&nummaxpoly
#define DENOM_VARS &denomtermlist,&denompoly,denomdegterms,&denomtoterms,&denommaxterms,&denommaxpoly
int nodemap(int node);
int internal_node(void);
void release(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly);
void evaluate_fn(int e,int a,int r, int sign, int numerator,int denominator);
void collect_tree(int r, int sign, int* tree);
void print_treebuf2(int n);
void treecheck(int r, int* d1, int* d2, int* tree, int* perm);
void product(int r, int* tree, int pass, int q);
void minty(int n, int e, int r, int pass, int f);
void init(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly);
void getexpr(int sign, int deg, int factors, int* factorlist, int edges,int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly);
void printexp(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly);
void makecancellationlist(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly);
int getsym(int i, int edges);

int main()
{
int i,j,k,scan, node1, node2, node3, node4, int1, int2, ch0, ch1, f;
char string[132], ch;
   printf("program file name tree29.c\n\r");

   for(;;){

     printf("\n\r\n\rinput data:\n\r\n\r");

#if dynamic
tbuf1max = 1000;
tbuf2max = 1000;
treebuf = (void*)malloc(tbuf1max*sizeof(array));
treebuf2 = (void*)malloc(tbuf2max*sizeof(array));
#endif

    test = 0;
    p = 1;			/* first edge is always dummy */
    branches = 1;
    n = 0;
    int_node = 999;
    a = 0;
    f = 0;


    for(i = 0; i < nmax; i++)nodelist[i] = 0;

    for (scan = 1; scan == 1;) {
	if(scanf("%s", &branch[branches + 1][0]) == EOF) return(0) ;
        printf("%s", &branch[branches + 1][0]);
        ch0 = branch[branches + 1][0];
        ch1 = branch[branches + 1][1];
        if( ch0 == '{'){
         scanf("%c",&ch);
         while( ch != '}'){
               printf("%c",ch);
               scanf("%c",&ch);
             };
          printf("}\n\r");
        }
       else
	if ( (((ch0 == 'g') || (ch0 == 'G') || (ch0 == 'r') || (ch0 == 'R') ) && ((ch1 == 'm') || (ch1 == 'M'))) ||
            (ch0 == 'o') || (ch0 == 'O') || (ch0 == 'u') || (ch0 == 'U') || (ch0 == 'b') || (ch0 == 'B') ){
	    scanf(" %u %u %u %u", &node1, &node2, &node3, &node4);
            printf(" %u %u %u %u\n\r", node1, node2, node3, node4);
         switch(ch0){
          case 'g':
          case 'G':
          case 'o':
          case 'O':
            branches++;
	    a++;
            a0[a] = branches;
	    a1[a] = nodemap(node1);
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node3);
            a4[a] = nodemap(node4);
           break;
          case 'r':
          case 'R':

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap(node3);
            p2[p] = internal_node();

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = internal_node();
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node4);
            a4[a] = p2[p];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap(node1);
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node2);
            a4[a] = a1[a-1];
            branch[branches][0] = 'o';
            branch[branches][1] = 0;


            branches++;
            a++;
            a0[a] = branches;
            a1[a] = p2[p];
            a2[a] = nodemap(node4);
            a3[a] = nodemap(node4);
            a4[a] = nodemap(node3);
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap(node1);
            p2[p] = a1[a-2];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

           break;

          case 'u':
          case 'U':
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = internal_node();
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node3);
            a4[a] = nodemap(node4);

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap(node1);
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node2);
            a4[a] = a1[a-1];
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = a1[a-1];
            p2[p] = nodemap(node1);
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

           break;

          case 'b':
          case 'B':
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap(node2);
            a2[a] = nodemap(node1);
            a3[a] = internal_node();
            a4[a] = nodemap(node4);

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = a3[a-1];
            a2[a] = nodemap(node4);
            a3[a] = nodemap(node4);
            a4[a] = nodemap(node3);
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap(node3);
            p2[p] = a3[a-1];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

           break;

        }
       }
         else
            if((ch0 == 'm') || (ch0 == 'M')){
            scanf(" %u %u %u %u", &node1, &node2, &node3, &node4);
            printf(" %u %u %u %u\n\r", node1, node2, node3, node4);

            int1 = internal_node();
            int2 = internal_node();

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap(node3);
            p2[p] = internal_node();
            k = branches;

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = internal_node();
            a2[a] = int2;
            a3[a] = int1;
            a4[a] = p2[p];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap(node1);
            a2[a] = int2;
            a3[a] = int2;
            a4[a] = a1[a-1];
            branch[branches][0] = 'o';
            branch[branches][1] = 0;


            branches++;
            a++;
            a0[a] = branches;
            a1[a] = p2[p];
            a2[a] = int1;
            a3[a] = int1;
            a4[a] = nodemap(node3);
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap(node1);
            p2[p] = a1[a-2];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;


/* second part */

            branches++;
            p++;
            for(i = 0;(branch[branches][i] = branch[k][i]) != 0; i++);
            p0[p] = branches;
            p1[p] = int2;
            p2[p] = internal_node();

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = internal_node();
            a2[a] = nodemap(node4);
            a3[a] = nodemap(node2);
            a4[a] = p2[p];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = int1;
            a2[a] = nodemap(node4);
            a3[a] = nodemap(node4);
            a4[a] = a1[a-1];
            branch[branches][0] = 'o';
            branch[branches][1] = 0;


            branches++;
            a++;
            a0[a] = branches;
            a1[a] = p2[p];
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node2);
            a4[a] = int2;
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = int1;
            p2[p] = a1[a-2];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;



              }
        else
        if ( ((ch0 == 'g') || (ch0 == 'G') ) && ((ch1 == 'y') || (ch1 == 'Y'))) {
            scanf(" %u %u %u %u", &node1, &node2, &node3, &node4);
            printf(" %u %u %u %u\n\r", node1, node2, node3, node4);

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap(node1);
            a2[a] = nodemap(node2);
            a3[a] = nodemap(node3);
            a4[a] = nodemap(node4);

            branches++;
            for(i = 0;(branch[branches][i] = branch[branches - 1][i]) != 0; i++);

            a++;
            a0[a] = branches;
            a1[a] = nodemap(node4);
            a2[a] = nodemap(node3);
            a3[a] = nodemap(node1);
            a4[a] = nodemap(node2);



        }
	else 
             if( (ch0 == 'f') || (ch0 == 'F')) {
              if(f == 1){
               printf("error-- more than one port indicator is used\n\r");
               return;
               };
               f = 1;
            scanf(" %u %u %u %u", &ni1, &nj1, &nk1, &nl1);
            printf(" %u %u %u %u\n\r", ni1, nj1, nk1, nl1);
               ni = nodemap(ni1);
               nj = nodemap(nj1);
               nk = nodemap(nk1);
               nl = nodemap(nl1);
              }
               else 
                  if( (ch0 == 'g') || (ch0 == 'G') ||
                    (ch0 == 'z') || (ch0 == 'Z') ||
                    (ch0 == 'y') || (ch0 == 'Y') ||
                    (ch0 == 'r') || (ch0 == 'R') ||
                    (ch0 == 'l') || (ch0 == 'L') ||
                    (ch0 == 'c') || (ch0 == 'C') ){
            scanf(" %u %u", &node1, &node2);
            printf(" %u %u\n\r", node1, node2);
            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap(node1);
            p2[p] = nodemap(node2);
                   }
             else
             if( (ch0 == 'e') || (ch0 == 'E')) {
/* dummy edge (handle) and extra node */
            p0[1] = 1;
            p1[1] = n + 1;
            p2[1] = 1;
            branch[1][0] = 'x';
            branch[1][1] = 0;
            n++;
            scan = 0;
        }
             else
             if( (ch0 == 'h') || (ch0 == 'H')) {
             printf("\n\rhelp:\n\r");
             printf("comments (on their own lines) are bracked:  {comments... }\n\r");
             printf("program written by Jeff Furman 5-4-2001\n\r");
             printf("one-port (two-terminal) passive elements:\n\r");
             printf("r,l,c,g,y,z (lower case or upper case) prefix, null up to 8 character suffix,\n\r");
             printf("whitespace, first node#(0-999),whitespace,second node#(0-999) examples--\n\r");
             printf("r 0 4\n\r");
             printf("L23a 7 41\n\r");
             printf("g2 3 8\n\r");
             printf("(gm, gy, and rm are reserved for two-port elements)\n\r");
             printf("two-port (four-terminal) active elements:\n\r");
             printf("gm,gy,o,u,b,rm,m (lower case or upper case) prefix, null up to 8 character suffix,\n\r");
             printf("whitespace, +output node#, whitespace, -output node#, whitespace, +input node#, whitespace, -input node#,\n\r");
             printf("positive current enters +nodes.\n\r");
             printf("gm denotes voltage controlled current source,\n\r");
             printf("o denotes ideal op amp,\n\r");
             printf("u (mu) denotes voltage controlled voltage source,\n\r");
             printf("b (beta) denotes current controlled current source,\n\r");
             printf("rm denotes current controlled voltage source,\n\r");
             printf("gy denotes a gyrator, and\n\r");
             printf("m denotes a mutual inductance\n\r");
             printf("examples--\n\r");
             printf("gm2 1 2 3 4\n\r");
             printf("U3 3 0 7 5\n\r");
             printf("o 6 3 2 9\n\r");
             printf("O 3 1 3 4\n\r");
             printf("f (lower case or upper case),\n\r");
             printf("whitespace, +port1 node#, whitespace, -port1 node#, whitespace, +port2 node#, whitespace, -port2 node#\n\r");
             printf("f 1 0 2 0\n\r");
             printf("e or end  terminates input\n\r");
             printf("t # before end line shows intermediate results: 1-input data dump,2-raw tree listing,4-buffered tree listing,\n\r8-matched trees and permutation listing,16-mixed unit trees,32-admittance trees,64-resource usage\n\r"); 
             printf("resource burden of input element types:\n\r");
             printf("elem prefix, branch name list, node list, passive edges, active edges\n\r");
             printf("r,l,c,g,y,z        1           no extra         1             0\n\r");
             printf(" GM, O             1           no extra         0             1\n\r");
             printf(" GY                2           no extra         0             2\n\r");
             printf("  U, B             3            1 extra         1             2\n\r");
             printf(" RM                5            2 extra         2             3\n\r");
             printf("  M               10            6 extra         4             6\n\r");
             printf(" the extra nodes are invisible externally\n\r");
             printf(" currently configured resource limits:\n\r");
             printf("branch name list %d\n\r",emax);
             printf("node list %d\n\r",nmax);
             printf("passive edges %d\n\r",pmax);
             printf("active edges %d\n\r",actmax);
#if dynamic
             printf("this version uses dynamic allocation for tree buffers treebuf and treebuf2\n\r");
#else
             printf("this version uses fixed allocation for tree buffers treebuf and treebuf2:\n\r");
             printf("treebuf allows %d trees; treebuf2 allows %d trees\n\r",tbuf1max,tbuf2max);
#endif
             printf("{sample sallen-key 3rd order lowpass:}\n\r");
             printf("{optional intermediate results}\n\r");
             printf("t 64\n\r");
             printf("r1 1 2\n\r");
             printf("r2 2 3\n\r");
             printf("r3 3 4\n\r");
             printf("c 2 5\n\r");
             printf("c 3 6\n\r");
             printf("c 4 5\n\r");
             printf("{vcvs here}\n\r");
             printf("u 6 5 4 5\n\r");
             printf("f 6 5 1 5\n\r");
             printf("end\n\r");
             return;
        }
             else
             if( (ch0 == 't') || (ch0 == 'T')) {
               scanf(" %u",&test);
               printf(" %u\n\r", test);
           }
           else

              {
                  printf("unrecognized element prefix-- '%s': must be r,l,c,g,y,z,gm,gy,o,u,b,rm,m,f,e\n\r",&branch[branches + 1][0]);
                  printf("h as input prints help\n\r");
                  return;
              };



        if(p >= pmax)
         {
           printf("number of one-port elements(= %d) exceeds limit of %d\n\r or improper input applied\n\r",p,pmax-1);
           return;
         };

        if(a >= actmax)
         {
           printf("number of two-port elements(= %d) exceeds limit of %d\n\r or improper input applied\n\r",a,actmax-1);
           return;
         };


    }
    printf("\n\r\n\r\n\r");

   if(f == 0){
             printf("please provide a port node indication: f (lower case or upper case),\n\r");
             printf("whitespace, +port1 node#, whitespace, -port1 node#, whitespace, +port2 node#, whitespace, -port2 node#\n\
r");
             printf("example:\n\r");
             printf("f 1 0 2 4\n\r");

   return;
   };
  if( (test & PRINT_STATS ) != 0){
             printf(" currently configured resource limits:\n\r");
             printf("branch name list limit:%d, used:%d\n\r",emax, branches);
             printf("node list limit:%d, used:%d\n\r",nmax, n);
             printf("passive edges limit:%d, used:%d\n\r",pmax, p);
             printf("active edges limit:%d, used:%d\n\r",actmax, a);
  };

    e = p;
    r = n - 1;


  if( (test & INPUT_DUMP ) != 0){
    printf("Internal representation of data including generated elements and reassigned node#s:\n");
    printf("edge,name,node#'s...\n\r");
    for (i = 1,j=1; i <= p; i++,j++) {
        printf("%3d %s %d %d\n", j, &branch[p0[i]][0], p1[i], p2[i]);
    };
    for (i = 1; i <= a; i++,j++) {
        printf("%3d %s %d %d %d %d\n", j, &branch[a0[i]][0], a1[i], a2[i], a3[i], a4[i]);
    };
        printf("%3d f %d %d %d %d\n\n", j, ni, nj, nk, nl);
        printf("internal node#  input node#:\n\r");
      for(i=1; i<=n; i++)
        printf("         %5d    %d\n\r",i, nodelist[i]);
      printf("\n\r\n\redge namelist: index, value\n\r");
      for(i=1; i <= branches; i++)
       printf("%3d, %s\n\r",i, &branch[i][0]);
      printf("\n\r\n\r");
   };

/* numerator and denominator indicators
#define V     0
#define wijkl 1
#define wijij 2
#define wklkl 3
#define wklij 4
#define uijkl 5
*/

/*
transfer ratios:
Vkl/Vij = wijkl/wijij
Ikl/Iij = -wijkl/wklkl
open circuit admittance functions:
Iij/Vij = V/wijij
Ikl/Vij = V/wklij
Iij/Vkl = V/wijkl
Ikl/Vkl = V/wklkl
short circuit admittance functions:
Iij/Vij = wklkl/uijkl
Iij/Vkl = -wklij/uijkl
Ikl/Vij = -wijkl/uijkl
Ikl/Vkl = wijij/uijkl
*/

#define ev_fn(s,n,d) evaluate_fn(e,a,r, s,n,d)

printf(" Function:\n\n V( %d, %d)/V( %d, %d) for I( %d, %d)=0\n",nk1,nl1,ni1,nj1,nk1,nl1);
ev_fn(1,wijkl,wijij);

printf(" Function:\n\n V( %d, %d)/V( %d, %d) for I( %d, %d)=0\n",ni1,nj1,nk1,nl1,ni1,nj1);
ev_fn(1,wklij,wklkl);

printf(" Function:\n\n I( %d, %d)/I( %d, %d) for V( %d, %d)=0\n",nk1,nl1,ni1,nj1,nk1,nl1);
ev_fn(-1,wijkl,wklkl);

printf(" Function:\n\n I( %d, %d)/I( %d, %d) for V( %d, %d)=0\n",ni1,nj1,nk1,nl1,ni1,nj1);
ev_fn(-1,wklij,wijij);

printf(" Function:\n\n Y( %d, %d; %d, %d) = I( %d, %d)/V( %d, %d) for V( %d, %d)=0\n"
       ,ni1,nj1,ni1,nj1,ni1,nj1,ni1,nj1,nk1,nl1);
ev_fn(1,wklkl,uijkl);

printf(" Function:\n\n Y( %d, %d; %d, %d) = I( %d, %d)/V( %d, %d) for V( %d, %d)=0\n"
       ,nk1,nl1,ni1,nj1,nk1,nl1,ni1,nj1,nk1,nl1);
ev_fn(-1,wijkl,uijkl);

printf(" Function:\n\n Y( %d, %d; %d, %d) = I( %d, %d)/V( %d, %d) for V( %d, %d)=0\n"
       ,ni1,nj1,nk1,nl1,ni1,nj1,nk1,nl1,ni1,nj1);
ev_fn(-1,wklij,uijkl);

printf(" Function:\n\n Y( %d, %d; %d, %d) = I( %d, %d)/V( %d, %d) for V( %d, %d)=0\n"
       ,nk1,nl1,nk1,nl1,nk1,nl1,nk1,nl1,ni1,nj1);
ev_fn(1,wijij,uijkl);

printf(" Function:\n\n Z( %d, %d; %d, %d) = V( %d, %d)/I( %d, %d) for I( %d, %d)=0\n"
       ,ni1,nj1,ni1,nj1,ni1,nj1,ni1,nj1,nk1,nl1);
ev_fn(1,wijij,V);

printf(" Function:\n\n Z( %d, %d; %d, %d) = V( %d, %d)/I( %d, %d) for I( %d, %d)=0\n"
       ,nk1,nl1,ni1,nj1,nk1,nl1,ni1,nj1,nk1,nl1);
ev_fn(1,wijkl,V);

printf(" Function:\n\n Z( %d, %d; %d, %d) = V( %d, %d)/I( %d, %d) for I( %d, %d)=0\n"
       ,ni1,nj1,nk1,nl1,ni1,nj1,nk1,nl1,ni1,nj1);
ev_fn(1,wklij,V);

printf(" Function:\n\n Z( %d, %d; %d, %d) = V( %d, %d)/I( %d, %d) for I( %d, %d)=0\n"
       ,nk1,nl1,nk1,nl1,nk1,nl1,nk1,nl1,ni1,nj1);
ev_fn(1,wklkl,V);


#if dynamic
  free(treebuf);
  free(treebuf2);
#endif

  };
}
int nodemap(int node)
{
int i;
if(node > 999){
 printf("node label %d must be less than 1000\n\r",node);
 return(1);
 };

for(i = 1; i <= n; i++){
  if(nodelist[i] == node)return(i);
};
        if(n >= nmax)
         {
           printf("number of nodes(= %d) exceeds limit of %d\n\r or improper input applied\n\r",n,nmax-1);
           return(1);
         };
n++;
nodelist[n] = node;
return(n);
}
int internal_node(void)
{
        if(n >= nmax)
         {
           printf("number of nodes(= %d) exceeds limit of %d\n\r or improper input applied\n\r",n,nmax-1);
           return(1);
         };

n++;
int_node++;
nodelist[n] = int_node;
return(n);
}
void evaluate_fn(int e,int a,int r, int sign, int numerator,int denominator)
{
int i, j, n,passive_edges,minty_nodes,actual_nodes,active_edges,numerator_form,denominator_form ;

opampcnt = 0;
minty_nodes = r + 1;
actual_nodes = r;
passive_edges = e;
    nbuf = 0;
    nbuf2 = 0;
    tbuf1overage = 0;
    tbuf2overage = 0;

    for (i = 0; i <= (e + a); i++)
        branchtag[i] = 0;

   for(i=0;i<maxsym;i++)
     cancelfactor[i] = 0x7fffffff;

  cancelcoeff = 0;
  cancelsign = -1;
  cancelexp = 0x7fffffff;

    for (i = 1; i <= e; i++) {
        d0[i] = p0[i];
        d1[i] = p1[i];
        d2[i] = p2[i];
    };

    for (i = 1; i <= a; i++) {
        d0[e + i] = a0[i];
        d1[e + i] = a1[i];
        d2[e + i] = a2[i];
    };



    switch(numerator){
    case V:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("V:\n\r");
          active_edges = a;
          numerator_form = 1; 
          for (i = 1; i <= a; i++) {
                j = d1[e + i];
                d1[e + i] = d2[e + i];
                d2[e + i] = j;
          };
          break;

    case wijkl:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wijkl:\n\r");
          active_edges = a + 1;
          numerator_form = 0;
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          d0[e+a+1] = branches+1;
          break;

    case wijij:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wijij:\n\r");
          active_edges = a + 1;
          numerator_form = 0;
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          break;

    case wklkl:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wklkl:\n\r");
          active_edges = a + 1;
          numerator_form = 0;
          d1[e+a+1] = nk;
          d2[e+a+1] = nl;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          break;

    case wklij:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wklij:\n\r");
          active_edges = a + 1;
          numerator_form = 0;
          d1[e+a+1] = nk;
          d2[e+a+1] = nl;
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          d0[e+a+1] = branches+1;          
          break;

    case uijkl:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("uijkl:\n\r");
          active_edges = a + 2;
          numerator_form = 2;
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = '1';
          branch[branches+1][2] = 0;
          d1[e+a+2] = nk;
          d2[e+a+2] = nl;
          d0[e+a+2] = branches+2;          
          branch[branches+2][0] = 'f';
          branch[branches+2][1] = '2';
          branch[branches+2][2] = 0;
          break;

    };

  for(i = 1; i <= (passive_edges+active_edges); i++){
   dv1[i] = d1[i];
   dv2[i] = d2[i];
  };

 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM  )) != 0){
printf("v-graph for numerator starts\n\r");
  };


    minty(minty_nodes,passive_edges+active_edges,actual_nodes, 
            vtree,numerator_form);


 if( (test & (PRINT_STATS )) != 0){

  printf("v-trees in treebuf limit %d, used %d\n\r",tbuf1max ,nbuf);

 };

    switch(numerator){
    case V:
          break;

    case wijkl:
          d1[e+a+1] = (sign == 1)?nk:nl;
          d2[e+a+1] = (sign == 1)?nl:nk;
          break;

    case wijij:
          if(sign == -1){
          j = d1[e+a+1];
          d1[e+a+1] = d2[e+a+1];
          d2[e+a+1] = j;
          };
          break;

    case wklkl:
          if(sign == -1){
          j = d1[e+a+1];
          d1[e+a+1] = d2[e+a+1];
          d2[e+a+1] = j;
          };
          break;

    case wklij:
          d1[e+a+1] = (sign == 1)?ni:nj;
          d2[e+a+1] = (sign == 1)?nj:ni;
          break;

    case uijkl:
          break;

    };

    for (i = 1; i <= a; i++) {
        d1[e + i] = a3[i];
        d2[e + i] = a4[i];
    };

 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)
printf("i-graph for numerator starts\n\r");

    minty(minty_nodes,passive_edges+active_edges,actual_nodes,
            itree,numerator_form);


    for (i = 1; i <= a; i++) {
        d1[e + i] = a1[i];
        d2[e + i] = a2[i];
    };


    switch(numerator){
    case V:
          break;

    case wijkl:
          break;

    case wijij:
          break;

    case wklkl:
          break;

    case wklij:
          break;

    case uijkl:
          break;

    };

    if(tbuf1overage != 0)printf("*******treebuf limit (tbuf1max= %d) exceeded by %d trees\n\r",tbuf1max, tbuf1overage);


    tbuf1overage = 0;

 if( (test & (PRINT_STATS)) != 0){
  printf("number of common (v- & i-) numerator trees in treebuf2= %d\n\r",nbuf2);
 };

    nbuf = 0;
    dtag = nbuf2 + 1;


    switch(denominator){
    case V:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("V:\n\r");
          active_edges = a;
          denominator_form = 1;
          break;

    case wijkl:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wijkl:\n\r");
          active_edges = a + 1;
          denominator_form = 0;
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          break;

    case wijij:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wijij:\n\r");
          active_edges = a + 1;
          denominator_form = 0;
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          break;

    case wklkl:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wklkl:\n\r");
          active_edges = a + 1;
          denominator_form = 0;
          d1[e+a+1] = nk;
          d2[e+a+1] = nl;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          break;


    case wklij:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("wklij:\n\r");
          active_edges = a + 1;
          denominator_form = 0;
          d1[e+a+1] = nk;
          d2[e+a+1] = nl;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = 0;
          break;

    case uijkl:
 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)printf("uijkl:\n\r");
          active_edges = a + 2;
          denominator_form = 2;
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          d0[e+a+1] = branches+1;          
          branch[branches+1][0] = 'f';
          branch[branches+1][1] = '1';
          branch[branches+1][2] = 0;
          d1[e+a+2] = nk;
          d2[e+a+2] = nl;
          d0[e+a+2] = branches+2;          
          branch[branches+2][0] = 'f';
          branch[branches+2][1] = '2';
          branch[branches+2][2] = 0;
          break;

    };

  for(i = 1; i <= (passive_edges+active_edges); i++){
   dv1[i] = d1[i];
   dv2[i] = d2[i];
  };


 if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)
printf("v-graph for denominator starts\n\r");


    minty(minty_nodes,passive_edges+active_edges,actual_nodes,
            vtree,denominator_form);


 if( (test & (PRINT_STATS)) != 0){

  printf("v-trees in treebuf limit %d, used %d\n\r",tbuf1max ,nbuf);

 };


    for (i = 1; i <= a; i++) {
        d1[e + i] = a3[i];
        d2[e + i] = a4[i];
    };


    switch(denominator){
    case V:
          break;

    case wijkl:
          d1[e+a+1] = nk;
          d2[e+a+1] = nl;
          break;

    case wijij:
          break;

    case wklkl:
          break;

    case wklij:
          d1[e+a+1] = ni;
          d2[e+a+1] = nj;
          break;

    case uijkl:
          break;

    };

if( (test & (PRINT_STATS | RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0)
printf("i-graph for denominator starts\n\r");

    minty(minty_nodes,passive_edges+active_edges,actual_nodes,
            itree,denominator_form);


    for (i = 1; i <= a; i++) {
        d1[e + i] = a3[i];
        d2[e + i] = a4[i];
    };


    switch(denominator){
    case V:
          for (i = 1; i <= a; i++) {
                j = d1[e + i];
                d1[e + i] = d2[e + i];
                d2[e + i] = j;
          };
          break;

    case wijkl:
          for (i = 1; i <= a; i++) {
                j = d1[e + i];
                d1[e + i] = d2[e + i];
                d2[e + i] = j;
          };
          break;

    case wijij:
          break;

    case wklkl:
          break;

    case wklij:
          break;

    case uijkl:
          break;

    };

    if(tbuf1overage != 0)printf("*******treebuf limit (tbuf1max= %d) exceeded by %d trees\n\r",tbuf1max, tbuf1overage);

    if(tbuf2overage != 0)printf("*******treebuf2 limit (tbuf2max= %d) exceeded by %d trees\n\r",tbuf2max, tbuf2overage);

 if( (test & (PRINT_STATS)) != 0){

  printf("number of common (v- & i-) denominator trees in treebuf2= %d\n\r",nbuf2-dtag+1);
  printf("total numerator and denominator trees in treebuf2= %d\n\r",nbuf2);
  printf("limit of sum is %d\n\r",tbuf2max);
  };

    print_treebuf2((e + a));

}

void collect_tree(int r, int sign, int* tree)
{
/* branchtag collects all the impedance-unit factors to be divided out later */
    int i, op, ch0;


    if (nbuf2 >= (tbuf2max - 1)) {
#if dynamic
tbuf2max = tbuf2max*2;
treebuf2 = (void*)realloc(treebuf2, tbuf2max*sizeof(array));
#else
	if(tbuf2overage == 0)printf("****treebuf2 is full\n");
        tbuf2overage++;
	return;
#endif
    };
    nbuf2++;
    treebuf2[nbuf2][0] = (sign == 1) ? r : (-r);
    for (i = 1, op = 0; i <= r; i++) {
        ch0 = branch[d0[tree[i]]][0];
	if ((ch0 == 'r') ||
	    (ch0 == 'R') ||
            (ch0 == 'z') ||
            (ch0 == 'Z') ||
	    (ch0 == 'l') ||
	    (ch0 == 'L') ||
	    (ch0 == 'm') ||
	    (ch0 == 'M'))
	    branchtag[d0[tree[i]]] = 1;
        if ((ch0 == 'o') ||
            (ch0 == 'O'))
          op++;
	treebuf2[nbuf2][i] = tree[i];
    };
    if( opampcnt < op)opampcnt = op;


 if( (test & ( BUFF_TREES )) != 0){
   printf("buffered trees:name[edge#]... ");
   for(i = 1;i <= r;i++) {
   printf(" %s[%d]",&branch[d0[tree[i]]][0],tree[i]);
   }
   printf("\n");
 };

}


void print_treebuf2(int n)
{
    int s, i, j, k, ch0, r, f, op, tmptree[emax], term[nmax];


 if( (test & (PRINT_STATS)) != 0)
    printf("opampcnt= %d\n\r",opampcnt);

   init(NUM_VARS);
   init(DENOM_VARS);

        if( (test & ( ADMITT_UNITS )) != 0)
         {
          printf("factors to be cancelled to form correct admittance products:\n\r");
        for (j = 1; j <= n; j++)
           {
            if (branchtag[j] == 1)
                printf(" %s[%d]", &branch[j][0] ,j);
           }; 
         printf("\n\r");
         };

if( (test & ( MIXED_UNITS | ADMITT_UNITS )) != 0)
  printf("numerator:\n\r");

    for (i = 1; i <= nbuf2; i++) {

if ( (i == dtag) )
if( (test & ( MIXED_UNITS | ADMITT_UNITS )) != 0)
  printf("denominator:\n\r");


	r = treebuf2[i][0];
	r = (r > 0) ? r : (-r);

/* opamp constraint */
        for (j = 1, op = 0; j <= r; j++)
            if( (branch[d0[treebuf2[i][j]]][0] == 'o') ||
                (branch[d0[treebuf2[i][j]]][0] == 'O') )
             op++;

        if( op != opampcnt) continue;


 if( (test & ( MIXED_UNITS )) != 0){
   printf("mixed units in tree admittance products: name[nameindex]...\n\r    ");
   for(j = 1;j <= r;j++) {
            ch0 = branch[d0[treebuf2[i][j]]][0];
            if (((ch0 == 'f') || (ch0 == 'F')) )
                continue;
            if (((ch0 == 'o') || (ch0 == 'O')))
                continue;
            /* elements of type 'a' are suppressed in all products,
               since they are considered to have unit values. */
            if (ch0 == 'a')
                continue;

   printf(" %s[%d]",&branch[d0[treebuf2[i][j]]][0],d0[treebuf2[i][j]]);
   }
   printf("\n");
 };
 
	for (j = 1; j <= n; j++)
	    tmptree[j] = 0;

	for (j = 1; j <= r; j++)
	    tmptree[d0[treebuf2[i][j]]] = 1;
/* (branchtag == tmptree == 0) means no factor in tree or cancellation list
   (branchtag == tmptree == 1) means the factor in tree is cancelled by cancellation list
   (branchtag != tmptree) means factor is due to either tree or cancellation list */
	for (j = 1; j <= n; j++)
	    if (branchtag[j] == tmptree[j])
		tmptree[j] = 0;
	    else
		tmptree[j] = 1;
	s = 0;
	f = 0;
	for (j = 1; j <= n; j++) {
	    if (tmptree[j] == 0)
		continue;
	    k = branch[j][0];
	    if (k == 'c')
		s++;
	    if (k == 'C')
		s++;
	    if (k == 'm')
		s++;
	    if (k == 'M')
		s++;
	    if (k == 'l')
		s++;
	    if (k == 'L')
		s++;
	} 

        if( (test & ( ADMITT_UNITS )) != 0)
                printf("admittance units for tree products: %s s%d", ((treebuf2[i][0]) > 0) ? "+" : "-", s);

	for (j = 1, k = 0; j <= n; j++) {
	    if (tmptree[j] == 0)
		continue;
            ch0 = branch[j][0];
	    if (((ch0 == 'f') || (ch0 == 'F')) )
		continue;
            if (((ch0 == 'o') || (ch0 == 'O')))
                continue;
            /* elements of type 'a' are suppressed in all products,
               since they are considered to have unit values. */
            if (ch0 == 'a')
                continue;
             term[k] = j;
            k++;

        if( (test & ( ADMITT_UNITS )) != 0)
	    printf(" %s[%d]", &branch[j][0], j);

	};

        if( (test & ( ADMITT_UNITS )) != 0)
         printf("\n");


       if (i >= dtag)
        getexpr( (((treebuf2[i][0]) > 0) ? 1 : -1), s, k, term, n, DENOM_VARS);
      else
        getexpr( (((treebuf2[i][0]) > 0) ? 1 : -1), s, k, term, n, NUM_VARS);

    }

    makecancellationlist(NUM_VARS);
    makecancellationlist(DENOM_VARS);

    printf("\n Numerator:\n\n");

    printexp(NUM_VARS);
   

    printf("\n Denominator:\n\n");

    printexp(DENOM_VARS);

   release(NUM_VARS);
   release(DENOM_VARS);

    printf("\n End: \n\n");

}

void treecheck(int r, int* d1, int* d2, int* tree, int* perm)
{
    int stack[nmax], tag[nmax];
    int depth, i, j, done, root_node, princip_node;

    for (i = 1; i <= r; i++) {
	tag[i] = 0;
    };

    perm[0] = 0;
    perm[1] = 0;
    stack[0] = 1;
    depth = 0;
    for (done = 0; done == 0;) {
	root_node = stack[depth];
	for (i = 1; i <= r; i++) {
	    if (tag[i] == 1)
		continue;
	    if (d1[tree[i]] == root_node) {
		tag[i] = 1;
		princip_node = d2[tree[i]];
		perm[princip_node] = tree[i];
		break;
	    };
	    if (d2[tree[i]] == root_node) {
		tag[i] = 1;
		princip_node = d1[tree[i]];
		perm[princip_node] =  (-tree[i]);
		break;
	    };
	};
	if (i <= r) {
	    depth++;
	    stack[depth] = princip_node;
	} else {
	    if (depth > 0)
		depth--;
	    else
		done = 1;
	};
    };
    return;
}



void product(int r, int* tree, int pass, int q)
{
    int i, j, k, m, n, s, f, f1, f2, sign, tmp, itreepermbuf[nmax], vtreepermbuf[nmax];


/* q == 0 rejects all trees without 'f' branch; q == 1 rejects all trees with 'f' branch;
   q == 2 rejects all trees without both 'f1' and 'f2 branches */


    f = 0;
    f1 = 0;
    f2 = 0;
    for (i = (r - 1); i >= 1; i--) {
        if ((branch[d0[tree[i]]][0] == 'f') ||
            (branch[d0[tree[i]]][0] == 'F')){
            if(q == 1) return;
            f = 1;
            if((branch[d0[tree[i]]][1] == '1'))
               f1 = 1;
             else {
            if((branch[d0[tree[i]]][1] == '2'))
               f2 = 1;
              };
        }
    }
     if (((q == 0) || (q == 1)) && (q == f))
            return;
     if ( (q == 2) && ( (f1 == 0) || (f2 == 0) ) )
           return;


    switch (pass) {
    case vtree:
	{
	    if (nbuf >= (tbuf1max - 1)) {
#if dynamic
tbuf1max = tbuf1max*2;
treebuf = (void*)realloc(treebuf, tbuf1max*sizeof(array));
#else
		if(tbuf1overage == 0)printf("****treebuf is full\n");
                tbuf1overage++;
		return;
#endif
	    };
	    for (i = 1; i < r; i++) {
		treebuf[nbuf][i] = tree[i];
	    };
	    nbuf++;
	};
	break;
    case itree:
	sign = 1;
	{
	    m = 0;
  for(i=0,j=nbuf-1;(nbuf>0)&&(j>=i)&&(m==0);){
                k = (i+j)/2;
		for (n = 1; n < r; n++) {
		    if (treebuf[k][n] > tree[n]) {
			j = k-1;
			m = 0;
			break;
		    } else {
                      if (treebuf[k][n] < tree[n]) {
			i = k+1;
			m = 0;
			break;
		       }else
		         m = 1;
                    };
		};
	    };
    if (m == 0) {
	return;
    };
    sign = 1;
    treecheck(r - 1, d1, d2, tree, &itreepermbuf[0]);
    treecheck(r - 1, dv1, dv2, tree, &vtreepermbuf[0]);

 if( (test & ( TREE_PERM )) != 0) {
 printf(" common tree edges:");
   for(i = 1;i < r;i++)printf(" %3d ",tree[i]);
   printf("\n");

   printf("v-tree permutation: ");
   for(i=0;i<=r;i++){
   printf("%3d ",vtreepermbuf[i]);
   };printf("\n");
   printf("i-tree permutation: ");
   for(i=0;i<=r;i++){
   printf("%3d ",itreepermbuf[i]);
   };printf("\n");
}

/* capture the principal node component of the sign computation */
	    for (i = 2; i <= r; i++) {
		if (vtreepermbuf[i] < 0) {
		    sign = sign * (-1);
		    vtreepermbuf[i] = (-vtreepermbuf[i]);
		};
		if (itreepermbuf[i] < 0) {
		    sign = sign * (-1);
		    itreepermbuf[i] = (-itreepermbuf[i]);
		};
	    };
/* capture the permutation component of the sign computation */
	    for (i = 2; i <= r; i++) {
		  {
		    if (vtreepermbuf[i] != itreepermbuf[i]) {
			sign = sign * (-1);
			for (j = i + 1; j <= r; j++) {
			    if (vtreepermbuf[j] == itreepermbuf[i]) {
				tmp = vtreepermbuf[j];
				vtreepermbuf[j] = vtreepermbuf[i];
				vtreepermbuf[i] = tmp;
				break;
			    };
			};
		    };
		};
	    };
 if( (test & ( TREE_PERM )) != 0)  printf("resulting sign= %2d\n",sign);
	};
	collect_tree(r - 1, sign, tree);
	break;
    };

}


void minty(int n, int e, int r, int pass, int f)
{
    int vd1, vd2, i, j, k, l, m, it, nt;
    int v[nmax], edge[emax], vstk[stk1][nmax], estk[stk1][emax], tree[nmax];


if( (test & (RAW_TREES | BUFF_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS )) != 0){
 printf("graph to generate trees:\n\rname:edge#,nameindex,node1,#node2#\n\r"); 
for(i = 1;i <= e; i++)printf("%5s: d0[%3d]=%3d, d1=%3d, d2=%3d\n\r",&branch[d0[i]][0],i,d0[i],d1[i],d2[i]);
printf("\n\r\n\r");
};

    for (k = 1; k <= e; k++)
	edge[k] = 1;
    for (k = 1; k <= n; k++)
	v[k] = 0;
    it = 0;
    i = 1;
    nt = 0;
  l15:
    for (k = 1; (k <= e); k++)
	if (edge[k] == 1)
	    goto l40;
  l25:
    if (it <= 1){

  if( (test & PRINT_STATS ) != 0){
    printf("total number of raw trees for this graph is %d\n\r",nt);
  };
	return;
 };
    it--;
    for (k = 1; k <= e; k++)
	edge[k] = estk[it][k];
    for (k = 1, i = 0; k <= n; k++) {
	v[k] = vstk[it][k];
	if (v[k] > i)
	    i = v[k];
    };
    i++;
    goto l15;

  l40:
    l = k;
    edge[l] = 0;
    m = 0;
    for (k = 1; k <= e; k++)
	if (edge[k] != 0)
	    m++;
    if (m >= (n - 1)) {

	for (k = 1; k <= e; k++)
	    estk[it][k] = edge[k];
	for (k = 1; k <= n; k++)
	    vstk[it][k] = v[k];
	it++;
    }
  l60:
    edge[l] = -1;
    vd1 = d1[l];
    vd2 = d2[l];

    if (v[vd1] == 0)
	goto l95;
    if (v[vd2] == 0)
	goto l105;

    l = v[vd1];
    m = v[vd2];

    if (l == m)
	goto l25;
    if (l > m) {
	k = m;
	m = l;
	l = k;
    };
    for (k = 1; k <= n; k++) {
	if (v[k] > m)
	    v[k]--;
	else if (v[k] == m)
	    v[k] = l;
    };
    i--;
    goto l110;
  l95:
    if (v[vd2] == 0)
	goto l100;
    v[vd1] = v[vd2];
    goto l110;
  l100:
    v[vd1] = i;
    v[vd2] = i;
    i++;
    goto l15;
  l105:
    v[vd2] = v[vd1];
  l110:
    for (k = 1; k <= n; k++)
	if (v[k] != 1)
	    goto l15;
    nt++;
    l = 1;
    for (k = 2; k <= e; k++)	/* dummy edge 1 is always present, so skip it */
	if (edge[k] == -1) {
	    tree[l] = k;
	    l++;
	};

 if( (test & (RAW_TREES)) != 0){
 printf(" tree no. %5d edges:",nt); 
   for(k = 1;k < r;k++)printf(" %3d ",tree[k]);
   printf("\n"); 
};

    product(r, tree, pass, f);
    goto l25;

}


/* code to combine like terms */

void init(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly)
{
	int i,j,k;
        int (*poly)[maxdegree], (*termlist)[maxsym+1];
        *maxtermlist = maxterm * maxdegree;
        *maxpoly = maxterm;
        *ptermlist = (void*)malloc((*maxtermlist)*sizeof(int)*(maxsym+1));
        *ppoly = (void*)malloc((*maxpoly)*sizeof(int)*maxdegree);

        poly = *ppoly;
        termlist = *ptermlist;

        for(j=0;j<maxdegree;j++)for(k=0;k<(*maxpoly);k++)poly[k][j] = 0;

	for(i=0;i<(*maxtermlist);i++)for(j=0;j<=maxsym;j++)termlist[i][j] = 0;
	for(j=0;j<maxdegree;j++)degterms[j] = 0;
        *toterms = 0;
        termlistoverage = 0;
}
void release(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly)
{
	free(*ptermlist);
        free(*ppoly);

}



void getexpr(int sign, int deg, int factors, int* factorlist, int edges,int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly)
{
	int i,j,k,l,m,n;
        int (*poly)[maxdegree], (*termlist)[maxsym+1];

        poly = *ppoly;
        termlist = *ptermlist;

		for(i=0;i<maxsym;i++)term[i]=0;

		for(j = 0;j < factors;j++)
		{
                        i = getsym(factorlist[j], edges);
                       term[i] = term[i] + 1;
		};


		if(degterms[deg] == 0)
		{
			poly[0][deg] = *toterms;
                        for(i=0;i<maxsym;i++)
                               termlist[*toterms][i] = term[i];
                        termlist[*toterms][maxsym] = sign;
#if dynamic
			if (*toterms >= (*maxtermlist)){
                           *maxtermlist = 2*(*maxtermlist);
                           *ptermlist = (void*)realloc((*ptermlist),(*maxtermlist)*sizeof(int)*(maxsym+1));
                           termlist = *ptermlist;
        for(i=((*maxtermlist)/2);i<(*maxtermlist);i++)for(j=0;j<=maxsym;j++)termlist[i][j] = 0;
                        }
                        (*toterms)++;
#else
			if (*toterms >= ((maxterm * maxdegree)-1)){
                           if(termlistoverage == 0)printf("****termlist is full\n\r");
                           termlistoverage++;
                        }
                        else
                           (*toterms)++;
#endif
                        degterms[deg] = 1;

		}
		else
		{
			for(m=0,n = 0;(m<degterms[deg]) && (n == 0);m++)
			{
                              k = poly[m][deg];
					for(i=0;i<maxsym;i++)
						if(termlist[k][i] != term[i])break;
					if(i == maxsym)
					{
                                                n = 1;
                                                termlist[k][maxsym] += sign;

					};
			};
			if (n == 0)
			{
				k = degterms[deg];
                                poly[k][deg] = *toterms;
                        for(i=0;i<maxsym;i++)
                               termlist[*toterms][i] = term[i];
                        termlist[*toterms][maxsym] = sign;
#if dynamic
			if (*toterms >= (*maxtermlist)){
                           *maxtermlist = 2*(*maxtermlist);
                           *ptermlist = (void*)realloc((*ptermlist),(*maxtermlist)*sizeof(int)*(maxsym+1));
                           termlist = *ptermlist;
        for(i=((*maxtermlist)/2);i<(*maxtermlist);i++)for(j=0;j<=maxsym;j++)termlist[i][j] = 0;
                        }
                        (*toterms)++;
#else
                        if (*toterms >= ((maxterm * maxdegree)-1)){
                           if(termlistoverage == 0)printf("****termlist is full\n\r");
                           termlistoverage++;
                        }
                        else
                           (*toterms)++;
#endif
                                degterms[deg]++;
#if dynamic
                        if(degterms[deg] >= (*maxpoly)){
                         *maxpoly = 2*(*maxpoly);
                         *ppoly = (void*)realloc((*ppoly),(*maxpoly)*sizeof(int)*maxdegree);
                         poly = *ppoly;
        for(j=0;j<maxdegree;j++)for(k=((*maxpoly)/2);k<(*maxpoly);k++)poly[k][j] = 0;
                          };
#else
#endif


			};
		};

}
void printexp(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly)
{
   int i,j,k,l,m,n;

        int (*poly)[maxdegree], (*termlist)[maxsym+1];
        poly = *ppoly;
        termlist = *ptermlist;


  if(termlistoverage != 0)printf("*******termlist size exceeded by %d terms\n\r",termlistoverage);

   if (cancelcoeff == 0) cancelcoeff = 1;

        if( (test & ( ADMITT_UNITS )) != 0){
            printf("common factors to cancel in all terms of both num and denom:\n\r");
            printf("%s%d ",cancelsign>0?"+":"-",cancelcoeff);
            for(k=0;k<maxsym;k++)
               if((cancelfactor[k]) > 0){
                if((cancelfactor[k]) > 1)
                 printf(" %s^%-2d",&branch[k][0],(cancelfactor[k]));
                else
                 printf(" %s   ",&branch[k][0]);
               };

            printf("\n\r");
        }; 



   for(i=0,n=0;i<maxdegree;i++)
	{
		if(degterms[i] > 0)
		{
                        for(j=0,k=0;(j<degterms[i]) && (k == 0);j++)
                        {
                             l = poly[j][i];
                             if((termlist[l][maxsym] != 0))k = 1;
 
                         };

                     if(k == 1){
                        n++;
			printf("\n\n + s^%-2d (\n",i-cancelexp);
			for(j=0;j<degterms[i];j++)
			{
                                l = poly[j][i];
                                if((termlist[l][maxsym] == 0))continue;
				printf("         ");
				{
					if(((cancelsign*termlist[l][maxsym]/cancelcoeff) > 1) ||
                                           ((cancelsign*termlist[l][maxsym]/cancelcoeff) < 0))
						printf("%3d",cancelsign*termlist[l][maxsym]/cancelcoeff);
					else
						printf("   "); 
				};
				for(k=0,m=0;k<maxsym;k++)
				{
					if((termlist[l][k] - cancelfactor[k]) > 0)
					{
                                                m = 1;
						if((termlist[l][k] - cancelfactor[k]) > 1)
							printf(" %s^%-2d",&branch[k][0],(termlist[l][k] - cancelfactor[k]));
						else
							printf(" %s   ",&branch[k][0]);
					};
				};
                                if( (m == 0) && ((cancelsign*termlist[l][maxsym]/cancelcoeff) == 1)) printf(" 1");
				printf("\n");
			};
			printf("        )");
                    };
		};
	};
        if(n == 0) printf("        0");
	printf("\n");
}
void makecancellationlist(int (**ptermlist)[maxsym+1], int (**ppoly)[maxdegree], int* degterms, int *toterms, int* maxtermlist, int* maxpoly)
{
   int i,j,k,l,m,n;

        int (*poly)[maxdegree], (*termlist)[maxsym+1];

        poly = *ppoly;
        termlist = *ptermlist;

   for(i=0;i<maxdegree;i++)
        {
                if(degterms[i] > 0)
                {
                        for(j=0,k=0;(j<degterms[i]) && (k == 0);j++)
                        {
                             l = poly[j][i];
                             if((termlist[l][maxsym] != 0))k = 1;

                         };

                     if(k == 1){
                        if(cancelexp > i) cancelexp = i;
                        for(j=0;j<degterms[i];j++)
                        {
                                l = poly[j][i];
                                if((termlist[l][maxsym] == 0))continue;

                                for(k=0;k<maxsym;k++)
                                {
                                        if((termlist[l][k] < cancelfactor[k]) ) cancelfactor[k] = termlist[l][k];
                                };

                                if((termlist[l][maxsym] > 0)) cancelsign = 1;

                                if(((n = termlist[l][maxsym]) < 0)) n = -n;
                                if ( (cancelcoeff > 0) && (cancelcoeff != n)) {
                                      if (cancelcoeff > n){
                                          m = n;
                                          n = cancelcoeff;
                                        }
                                      else
                                          m = cancelcoeff;

                                      for(;n>m;)
                                       {
                                          for(;n>m;) n -= m;
                                           l = m;
                                           m = n;
                                           n = l;
                                        };
                                      cancelcoeff = n;
                                  }
                                else
                                 cancelcoeff = n;
                        };
                    };
                };
        };
}


int getsym(int i, int edges)
{

	int j,k,l,m;
	k = 1;
	while(k <= edges )
	{
		m = 0;
		while ((branch[k][m] == branch[i][m]) && (branch[i][m] != 0))m++;
		if (branch[k][m] == branch[i][m]) { return k;};
		k++;
	};

return i;

}

 
  
