 /* tree40ab.c
    maeyeda's 2-graph method to evaluate topological formulas using
    (minty's algorithm for spanning tree enumeration
in symbolic network analysis, lin) */
#define gcc 0
#if gcc
#include <stdio.h>
#include <time.h>
#else
#include "..\include\time.h"
#include "..\include\stdio.h"
#endif

#define ABS(x) (((x) >= 0)?(x):(-(x)))

/*resource burden for s and F parameter embedding network*/
#define s_active 4
#define s_passive 8
#define s_nodes 8

#define actmax (30+s_active)               /* max active 2-ports */
#define pmax (60+s_passive)                 /*max passive (one-port) edges */
#define emax (pmax+actmax)	/* max edges */
#define nmax (35+s_nodes)                 /* max nodes */
#define stk1 nmax		/* minty alg. stack */
#define stk2 nmax		/* traverse stack */

#define BRLEN 12 /*max chars in branch names*/

/* numerator and denominator indicators */
#define V     0
#define wijkl 1
#define wijij 2
#define wklkl 3
#define wklij 4
#define uijkl 5
#define numerator 6
#define denominator 7

/* test parameter masks */
#define INPUT_DUMP 1
#define RAW_TREES 2

#define TREE_PERM 8
#define MIXED_UNITS 16
#define ADMITT_UNITS 32
#define PRINT_STATS 64
#define DEBUG_128 128
#define DEBUG_256 256
#define DEBUG_512 512
#define DEBUG_1024 1024
#define DEBUG_2048 2048

#define OP 1
#define F1 2
#define F2 4
#define ONETREES 0
#define TWOTREES 2
#define THREETREES 6
#define TREEMASK 6
#define UNITY 8

int denfn, numfn, inptno, evfn, numcntxt;

int nbuf, nbuf1;
int branchtag[emax], impedancelist[emax], impedancecount;
int powermap[emax], dpowermap[emax], fmap[emax], dmap[emax];
int dtag;
int ni, nj, nk, nl, ni1, nj1, nk1, nl1, normalization, opampcnt, opamp, n, int_node, nodelist[nmax];
int ioprej, voprej;
int p0[pmax], p1[pmax], p2[pmax];
int test;

int e, p, r, a, d0[emax], di1[emax], di2[emax], dv1[emax], dv2[emax];
int a0[actmax], a1[actmax], a2[actmax], a3[actmax], a4[actmax];
volatile int branches;
char branch[emax][BRLEN];


#define maxsym emax
#define maxterm emax
#define maxdegree nmax

typedef struct vardef{
int (*termlist)[maxsym + 2];
int (*poly)[maxdegree];
int degterms[maxdegree];
int maxterms;
int maxpoly;
int toterms;
int opamps;
 } vardef;


vardef vars[8];

int topological_fun[6];

int term[maxsym], cancelfactor[maxsym], cancelcoeff, cancelsign, cancelexp;
int branchmap[maxsym], dbranchmap[maxsym];

int iit, i_nt, ivstk[stk1][nmax], iestk[stk1][emax], istate, itree[nmax];
int vit, vnt, vvstk[stk1][nmax], vestk[stk1][emax], vstate, vtree[nmax];
int mn, me, mr, mq;

clock_t start_time;


char* topo_names[] = {
   "V",
   "Wijkl",
   "Wijij",
   "Wklkl",
   "Wklij",
   "Uijkl",
};

typedef  struct f_tab{
   int fn_no;
   char* fn_name;
   int fn_p1p;
   int fn_p1m;
   int fn_p2p;
   int fn_p2m;
   char* var1_name;
   int var1_p1p;
   int var1_p1m;
   char* var2_name;
   int  var2_p1p;
   int  var2_p1m;
   char* var3_name;
   int  var3_p1p;
   int  var3_p1m;
   int sign;
   int num_form;
   int denom_form;
}f_tab;

f_tab function_table[] = {
   {20,"g",0,1,0,1,"i",0,1,"v",0,1,"i",2,3, 1,V,wijij},
   {3 ,"g",0,1,2,3,"i",0,1,"i",2,3,"v",0,1,-1,wklij,wijij},
   {0 ,"g",2,3,0,1,"v",2,3,"v",0,1,"i",2,3,-1,wijkl,wijij},
   {21,"g",2,3,2,3,"v",2,3,"i",2,3,"v",0,1, 1,uijkl,wijij},
   {22,"h",0,1,0,1,"v",0,1,"i",0,1,"v",2,3, 1,uijkl,wklkl},
   {1 ,"h",0,1,2,3,"v",0,1,"v",2,3,"i",0,1, 1,wklij,wklkl},
   {2 ,"h",2,3,0,1,"i",2,3,"i",0,1,"v",2,3,-1,wijkl,wklkl},
   {23,"h",2,3,2,3,"i",2,3,"v",2,3,"i",0,1, 1,V,wklkl},
   {4 ,"y",0,1,0,1,"i",0,1,"v",0,1,"v",2,3, 1,wklkl,uijkl},
   {6 ,"y",0,1,2,3,"i",0,1,"v",2,3,"v",0,1,-1,wklij,uijkl},
   {5 ,"y",2,3,0,1,"i",2,3,"v",0,1,"v",2,3,-1,wijkl,uijkl},
   {7 ,"y",2,3,2,3,"i",2,3,"v",2,3,"v",0,1, 1,wijij,uijkl},
   {8 ,"z",0,1,0,1,"v",0,1,"i",0,1,"i",2,3, 1,wijij,V},
   {10,"z",0,1,2,3,"v",0,1,"i",2,3,"i",0,1, 1,wklij,V},
   {9 ,"z",2,3,0,1,"v",2,3,"i",0,1,"i",2,3, 1,wijkl,V},
   {11,"z",2,3,2,3,"v",2,3,"i",2,3,"i",0,1, 1,wklkl,V},
   {24,"A",2,3,0,1,"v",0,1,"v",2,3,"i",2,3, 1,wijij,wijkl},
   {25,"B",2,3,0,1,"v",0,1,"-i",2,3,"v",2,3, 1,uijkl,wijkl},
   {26,"C",2,3,0,1,"i",0,1,"v",2,3,"i",2,3, 1,V,wijkl},
   {27,"D",2,3,0,1,"i",0,1,"-i",2,3,"v",2,3, 1,wklkl,wijkl},
   {28,"a",0,1,2,3,"v",2,3,"v",0,2,"i",0,1, 1,wklkl,wklij},
   {29,"b",0,1,2,3,"v",2,3,"-i",0,1,"v",0,1, 1,uijkl,wklij},
   {30,"c",0,1,2,3,"i",2,3,"v",0,1,"i",0,1, 1,V,wklij},
   {31,"d",0,1,2,3,"i",2,3,"-i",0,1,"v",0,1, 1,wijij,wklij},
/* the following functions use an embedding network for analysis*/
   {12,"s",0,1,0,1," ",0,0," ",0,0," ",0,0,-1,wklkl,uijkl},
   {13,"s",0,1,2,3," ",0,0," ",0,0," ",0,0, 1,wklij,uijkl},
   {14,"s",2,3,0,1," ",0,0," ",0,0," ",0,0, 1,wijkl,uijkl},
   {15,"s",2,3,2,3," ",0,0," ",0,0," ",0,0,-1,wijij,uijkl},
   {16,"F",0,1,0,1," ",0,0," ",0,0," ",0,0,-1,V,wijkl},
   {17,"F",0,1,2,3," ",0,0," ",0,0," ",0,0,-1,wklkl,wijkl},
   {18,"F",2,3,0,1," ",0,0," ",0,0," ",0,0, 1,wijij,wijkl},
   {19,"F",2,3,2,3," ",0,0," ",0,0," ",0,0, 1,uijkl,wijkl},
};

char* par_names[2][(sizeof(function_table)/sizeof(f_tab))] = {
   {"g11","g12","g21","g22","h11","h12","h21","h22",
      "y11","y12","y21","y22","z11","z12","z21","z22",
      "A","B","C","D","a","b","c","d",
      "s11","s12","s21","s22","f11","f12","f21","f22"},
   {"G11","G12","G21","G22","H11","H12","H21","H22",
      "Y11","Y12","Y21","Y22","Z11","Z12","Z21","Z22",
      "A21","B21","C21","D21","a12","b12","c12","d12",
      "S11","S12","S21","S22","F11","F12","F21","F22" }
};

int par_list[(sizeof(function_table)/sizeof(f_tab))];

#define VARLIST(x) & vars[x].termlist,& vars[x].poly, vars[x].degterms,& vars[x].toterms,& vars[x].maxterms,& vars[x].maxpoly

#define NUM_VARS VARLIST(numerator)
#define DENOM_VARS VARLIST(denominator)


#define PARAMETERLIST int (**ptermlist)[maxsym + 2], int (**ppoly)[maxdegree], int *degterms, int *toterms, int *maxtermlist, int *maxpoly

int nodemap (int node);
int internal_node (void);
void release (PARAMETERLIST);
void evaluate_fn (int e, int a, int r, int topofun);
#define ev_fn(n) evaluate_fn(e,a,r, n)


void release_topo_data(void);
void analyze_network(void);
int parse_input(void);
void embed_network_for_s_parameters(void);
int print_treebuf (int sign, int num_fun, int denom_fun);
void treecheck (int r, int *d1, int *d2, int *tree, int *perm);
int product (int minty_nodes, int edges, int actual_nodes, int form, int br, int topofun);
void iminty (void);
void vminty (void);
void combineinit (void);
void init (PARAMETERLIST);
int resize(void ** ptr, int * count, int module);
int copy(int sign, int dest, int source);
int getexpr (int sign, int deg, int factors, int *factorlist, int edges, PARAMETERLIST);
void printexp (PARAMETERLIST);
void makecancellationlist (PARAMETERLIST);
int getsym (int i);
int correctunits (PARAMETERLIST);
int fixup (void);
void eval_branchtag (PARAMETERLIST);
void eval_maps(void);

int 
main ()
{

   start_time = clock();

   printf ("program file name tree40ab.c\n");
   inptno = 0;

   for(;0 == parse_input();)analyze_network();

   printf("elapsed time: %d/%d sec.\n",(clock() - start_time),CLOCKS_PER_SEC);

   return(0);
}


int parse_input(void)
{
   int i, j, k, scan, node1, node2, node3, node4, int1, int2, ch0, ch1,
     f, par_flag;
   char ch, string[128];
   int port_node[4] = {1,1,2,2};

   printf ("\ninput data:\n");

   inptno++;

   test = 0;
   normalization = 0;        
   opampcnt = 0;
   p = 1;			/* first edge is always dummy */
   branches = 1;
   n = 0;
   int_node = 999;
   a = 0;
   f = 0;


   for (i = 0; i < nmax; i++)
      nodelist[i] = 0;

   for(i = 0; i < (sizeof(function_table)/sizeof(f_tab)); i++)
      par_list[i] = 1;

   par_flag = 0;

   for (scan = 1; scan == 1;) {
      branch[branches + 1][BRLEN - 1] = 0;
      if (scanf ("%s", &branch[branches + 1][0]) == EOF)
         return (1);
      if (branch[branches + 1][BRLEN - 1] != 0)
        {
        printf("following branch name is truncated to %d characters:\n",BRLEN - 1);
        branch[branches + 1][BRLEN - 1] = 0;
        };
      printf ("%s", &branch[branches + 1][0]);
      ch0 = branch[branches + 1][0];
      ch1 = branch[branches + 1][1];
      if (ch0 == '{') /* help vi-- }*/{
         scanf ("%c", &ch);
         while /*help vi-- {*/(ch != '}') {
            printf ("%c", ch);
            scanf ("%c", &ch);
         };
         printf /*help vi-- {*/("}\n");
      } else if ((((ch0 == 'g') || (ch0 == 'G') || (ch0 == 'r') || (ch0 == 'R')) && ((ch1 == 'm') || (ch1 == 'M'))) ||
        (ch0 == 'o') || (ch0 == 'O') || (ch0 == 'u') || (ch0 == 'U') || (ch0 == 'b') || (ch0 == 'B')) {
         scanf (" %u %u %u %u", &node1, &node2, &node3, &node4);
         printf (" %u %u %u %u\n", node1, node2, node3, node4);
         switch (ch0) {
         case 'o':
         case 'O':
            opampcnt++;
         case 'g':
         case 'G':
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap (node1);
            a2[a] = nodemap (node2);
            a3[a] = nodemap (node3);
            a4[a] = nodemap (node4);
            break;
         case 'r':
         case 'R':

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap (node3);
            p2[p] = internal_node ();

            branches++;
            a++;
            a0[a] = branches;
            a1[a] = internal_node ();
            a2[a] = nodemap (node2);
            a3[a] = nodemap (node4);
            a4[a] = p2[p];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            opampcnt++;
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap (node1);
            a2[a] = nodemap (node2);
            a3[a] = nodemap (node2);
            a4[a] = a1[a - 1];
            branch[branches][0] = 'o';
            branch[branches][1] = 0;


            opampcnt++;
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = p2[p];
            a2[a] = nodemap (node4);
            a3[a] = nodemap (node4);
            a4[a] = nodemap (node3);
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap (node1);
            p2[p] = a1[a - 2];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            break;

         case 'u':
         case 'U':
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = internal_node ();
            a2[a] = nodemap (node2);
            a3[a] = nodemap (node3);
            a4[a] = nodemap (node4);

            opampcnt++;
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap (node1);
            a2[a] = nodemap (node2);
            a3[a] = nodemap (node2);
            a4[a] = a1[a - 1];
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = a1[a - 1];
            p2[p] = nodemap (node1);
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            break;

         case 'b':
         case 'B':
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = nodemap (node2);
            a2[a] = nodemap (node1);
            a3[a] = internal_node ();
            a4[a] = nodemap (node4);

            opampcnt++;
            branches++;
            a++;
            a0[a] = branches;
            a1[a] = a3[a - 1];
            a2[a] = nodemap (node4);
            a3[a] = nodemap (node4);
            a4[a] = nodemap (node3);
            branch[branches][0] = 'o';
            branch[branches][1] = 0;

            branches++;
            p++;
            p0[p] = branches;
            p1[p] = nodemap (node3);
            p2[p] = a3[a - 1];
            branch[branches][0] = 'a';
            branch[branches][1] = 0;

            break;

         }
      } else if ((ch0 == 'm') || (ch0 == 'M')) {
         scanf (" %u %u %u %u", &node1, &node2, &node3, &node4);
         printf (" %u %u %u %u\n", node1, node2, node3, node4);

         int1 = internal_node ();
         int2 = internal_node ();

         branches++;
         p++;
         p0[p] = branches;
         p1[p] = nodemap (node3);
         p2[p] = internal_node ();
         k = branches;

         branches++;
         a++;
         a0[a] = branches;
         a1[a] = internal_node ();
         a2[a] = int2;
         a3[a] = int1;
         a4[a] = p2[p];
         branch[branches][0] = 'a';
         branch[branches][1] = 0;

         opampcnt++;
         branches++;
         a++;
         a0[a] = branches;
         a1[a] = nodemap (node1);
         a2[a] = int2;
         a3[a] = int2;
         a4[a] = a1[a - 1];
         branch[branches][0] = 'o';
         branch[branches][1] = 0;


         opampcnt++;
         branches++;
         a++;
         a0[a] = branches;
         a1[a] = p2[p];
         a2[a] = int1;
         a3[a] = int1;
         a4[a] = nodemap (node3);
         branch[branches][0] = 'o';
         branch[branches][1] = 0;

         branches++;
         p++;
         p0[p] = branches;
         p1[p] = nodemap (node1);
         p2[p] = a1[a - 2];
         branch[branches][0] = 'a';
         branch[branches][1] = 0;


/* second part */

         branches++;
         p++;
         for (i = 0; (branch[branches][i] = branch[k][i]) != 0; i++);
         p0[p] = branches;
         p1[p] = int2;
         p2[p] = internal_node ();

         branches++;
         a++;
         a0[a] = branches;
         a1[a] = internal_node ();
         a2[a] = nodemap (node4);
         a3[a] = nodemap (node2);
         a4[a] = p2[p];
         branch[branches][0] = 'a';
         branch[branches][1] = 0;

         opampcnt++;
         branches++;
         a++;
         a0[a] = branches;
         a1[a] = int1;
         a2[a] = nodemap (node4);
         a3[a] = nodemap (node4);
         a4[a] = a1[a - 1];
         branch[branches][0] = 'o';
         branch[branches][1] = 0;


         opampcnt++;
         branches++;
         a++;
         a0[a] = branches;
         a1[a] = p2[p];
         a2[a] = nodemap (node2);
         a3[a] = nodemap (node2);
         a4[a] = int2;
         branch[branches][0] = 'o';
         branch[branches][1] = 0;

         branches++;
         p++;
         p0[p] = branches;
         p1[p] = int1;
         p2[p] = a1[a - 2];
         branch[branches][0] = 'a';
         branch[branches][1] = 0;



      } else if (((ch0 == 'g') || (ch0 == 'G')) && ((ch1 == 'y') || (ch1 == 'Y'))) {
         scanf (" %u %u %u %u", &node1, &node2, &node3, &node4);
         printf (" %u %u %u %u\n", node1, node2, node3, node4);

         branches++;
         a++;
         a0[a] = branches;
         a1[a] = nodemap (node1);
         a2[a] = nodemap (node2);
         a3[a] = nodemap (node3);
         a4[a] = nodemap (node4);

         branches++;
         for (i = 0; (branch[branches][i] = branch[branches - 1][i]) != 0; i++);

         a++;
         a0[a] = branches;
         a1[a] = nodemap (node4);
         a2[a] = nodemap (node3);
         a3[a] = nodemap (node1);
         a4[a] = nodemap (node2);



      } else if ((ch0 == 'f') || (ch0 == 'F')) {
         if (f == 1) {
            printf ("error-- more than one port indicator is used\n");
            return (1);
         };
         f = 1;
         scanf (" %u %u %u %u", &ni1, &nj1, &nk1, &nl1);
         printf (" %u %u %u %u\n", ni1, nj1, nk1, nl1);
         ni = nodemap (ni1);
         nj = nodemap (nj1);
         nk = nodemap (nk1);
         nl = nodemap (nl1);
      } else if ((ch0 == 'p') || (ch0 == 'P')) {
         if (par_flag == 0) {
            par_flag = 1;
            for(i = 0; i < (sizeof(function_table)/sizeof(f_tab)); i++)
               par_list[i] = 0;
         };
         scanf (" %s",string);
         printf (" %s\n", string);

         i = 1;
         while (i < (sizeof(function_table)/sizeof(f_tab))) {
            j = 0;
            while ((string[j] == par_names[0][i][j]) && (par_names[0][i][j] != 0))
               j++;
            if (string[j] == par_names[0][i][j]) {
               par_list[i] = 1;
               break;
            };
            i++;
         };
         if(i == (sizeof(function_table)/sizeof(f_tab))){

            i = 1;
            while (i < (sizeof(function_table)/sizeof(f_tab))) {
               j = 0;
               while ((string[j] == par_names[1][i][j]) && (par_names[1][i][j] != 0))
                  j++;
               if (string[j] == par_names[1][i][j]) {
                  par_list[i] = 1;
                  break;
               };
               i++;
            };
            if(i == (sizeof(function_table)/sizeof(f_tab)))
               printf(" %s does not match the list of parameter names, see help info.\n",string);


         };


      } else if ((ch0 == 'g') || (ch0 == 'G') ||
        (ch0 == 'z') || (ch0 == 'Z') ||
        (ch0 == 'y') || (ch0 == 'Y') ||
        (ch0 == 'r') || (ch0 == 'R') ||
        (ch0 == 'l') || (ch0 == 'L') ||
        (ch0 == 'c') || (ch0 == 'C')) {
         scanf (" %u %u", &node1, &node2);
         printf (" %u %u\n", node1, node2);
         branches++;
         p++;
         p0[p] = branches;
         p1[p] = nodemap (node1);
         p2[p] = nodemap (node2);
      } else if ((ch0 == 'n') || (ch0 == 'N')) {
         scanf (" %u", &normalization);
         printf (" %u\n", normalization);
      } else if ((ch0 == 'e') || (ch0 == 'E')) {
/* dummy edge (handle) and extra node */
         p0[1] = 1;
         p1[1] = n + 1;
         p2[1] = 1;
         branch[1][0] = 'x';
         branch[1][1] = 0;
         n++;
         scan = 0;
      } else if ((ch0 == 'h') || (ch0 == 'H')) {
         printf ("\nhelp:\n");
         printf ("comments (on their own lines) are bracketed:  {comments... }\n");
         printf ("program written by Jeff Furman 5-4-2001 vers 40a\n");
         printf ("one-port (two-terminal) passive elements:\n");
         printf ("r,l,c,g,y,z (lower case or upper case) prefix, null up to 8 character suffix,\n");
         printf ("whitespace, first node#(0-999),whitespace,second node#(0-999) examples--\n");
         printf ("r 0 4\n");
         printf ("L23a 7 41\n");
         printf ("g2 3 8\n");
         printf ("(gm, gy, and rm are reserved for two-port elements)\n");
         printf ("two-port (four-terminal) active elements:\n");
         printf ("gm,gy,o,u,b,rm,m (lower case or upper case) prefix, null up to 8 character suffix,\n");
         printf ("whitespace, +output node#, whitespace, -output node#, whitespace, +input node#, whitespace, -input node#,\n");
         printf ("positive current enters +nodes except port 2 in ABCD parameters and port 1 in abcd parameters.\n");
         printf ("gm denotes voltage controlled current source,\n");
         printf ("o denotes ideal op amp,\n");
         printf ("u (mu) denotes voltage controlled voltage source,\n");
         printf ("b (beta) denotes current controlled current source,\n");
         printf ("rm denotes current controlled voltage source,\n");
         printf ("gy denotes a gyrator, and\n");
         printf ("m denotes a mutual inductance\n");
         printf ("examples--\n");
         printf ("gm2 1 2 3 4\n");
         printf ("U3 3 0 7 5\n");
         printf ("o 6 3 2 9\n");
         printf ("O 3 1 3 4\n");
         printf ("floating or isolated network regions must be avoided-- they\n");
         printf ("exist when only two-port elements connect them to the remainder of the\n");
         printf ("network. a remedy is sharing a common node (preferred, since it reduces the\n");
         printf ("graph complexity), or connecting these regions with single branches\n");
         printf ("these single branches eventually cancel, so they don't appear in the results,\n");
         printf ("but, they take their toll in increasing computation time. a rule of thumb is\n");
         printf ("all nodes must be connected in a single graph, when only one-port or\n");
         printf ("passive elements are considered.\n");
         printf ("mandatory port designator-- f <+p1> <-p1> <+p2> <-p2>:\n");
         printf ("f (lower case or upper case),\n");
         printf ("whitespace, +port1 node#, whitespace, -port1 node#, whitespace, +port2 node#, whitespace, -port2 node#\n");
         printf ("f 1 3 2 4\n");
         printf ("optional parameter designator-- may be repeated for multiple parameters;\n");
         printf ("if it is omitted, default is evaluating all 32 parameter functions:\n");
         printf ("p (lower or upper case), whitespace, one of these parameter names:\n");
         printf ("g11 g12 g21 g22 h11 h12 h21 h22 y11 y12 y21 y22 z11 z12 z21 z22\n");
         printf ("A B C D a b c d s11 s12 s21 s22 f11 f12 f21 f22\n");
         printf ("case of letters is sensitive for ABCD and abcd only\n");
         printf ("p g21\n");
         printf ("P y11\n");
         printf ("p Y12\n");
         printf ("p A\n");
         printf ("n # or N # specifies normalization for s and F parameters: 0(or absent)- Y0, 1- Z0, 2(or greater)- unity\n");
         printf ("e or end  terminates input\n");
         printf ("t # before end line shows intermediate results: 1-input data dump,2-raw tree listing,\n8-matched trees and permutation listing,16-mixed unit trees,32-admittance trees,64-resource usage\n");
         printf ("resource burden of input element types:\n");
         printf ("elem prefix, branch name list, node list, passive edges, active edges\n");
         printf ("r,l,c,g,y,z        1           no extra         1             0\n");
         printf (" GM, O             1           no extra         0             1\n");
         printf (" GY                2           no extra         0             2\n");
         printf ("  U, B             3            1 extra         1             2\n");
         printf (" RM                5            2 extra         2             3\n");
         printf ("  M               10            6 extra         4             6\n");
         printf ("any s or F param  12            8 extra         8             4\n");
         printf (" the extra nodes and branches are invisible in the output\n");
         printf ("\n");
         printf ("two-port parameter notation and equivalences:\n");
         printf ("for ports expressed as f <i> <j> <k> <l>\n");
         printf ("where <i> <j> <k> <l> are numeric node numbers labeling the port terminals,\n");
         printf ("then port 1 is denoted by (i,j) and port 2 is denoted by (k,l)\n");
         printf ("examples: y(i,j;k,l) == y12\n");
         printf ("g(k,l;i,j) == g21 etc.\n");
         printf ("concrete example: for\n");
         printf ("f 3 7 9 10\n");
         printf ("p g21\n");
         printf ("this would be:\n");
         printf ("g(9, 10; 3, 7) = v(9, 10)/v(3, 7) for i(9, 10) = 0\n");
         printf ("y denotes short circuit admittance parameters,\n");
         printf ("z denotes open  circuit impedance  parameters,\n");
         printf ("h denotes hybrid parameters,\n");
         printf ("g denotes inverse hybrid parameters,\n");
         printf ("A,B,C,D denotes forward chain or transmission parameters,\n");
         printf ("a,b,c,d denotes reverse chain parameters,\n");
         printf ("s denotes scattering parameters,\n");
         printf ("F denotes scattering forward chain parameters,\n");
         printf ("these parameters are directly computed:\n");
         printf ("g11, g12, g21, g22, h11, h12, h21, h22,\n");
         printf ("y11, y12, y21, y22, z11, z12, z21, z22,\n");
         printf ("A, B, C, D, a, b, c, d,\n");
         printf ("s11, s12, s21, s22, F11, F12, F21, F22.\n");
         printf ("parameter summary:\n");
         for(i=0;i<(sizeof(function_table)/sizeof(f_tab));i++){
            evfn = function_table[i].fn_no;
            if((evfn >=12) && (evfn <= 19)) continue; /*bypass s and F parameters*/

            printf (" %s%1d%1d = %s%1d/%s%1d for %s%1d = 0; topological formula: %c%s/%s\n",
              function_table[i].fn_name,
              port_node[function_table[i].fn_p1p] ,
              port_node[function_table[i].fn_p2p] ,
              function_table[i].var1_name,
              port_node[function_table[i].var1_p1p] ,
              function_table[i].var2_name,
              port_node[function_table[i].var2_p1p] ,
              function_table[i].var3_name,
              port_node[function_table[i].var3_p1p] ,
              (function_table[i].sign < 0)?'-':' ',
              topo_names[function_table[i].num_form] ,
              topo_names[function_table[i].denom_form] );

         };


         printf ("\n");
         printf ("references:\n");
         printf ("Huelsman, L. P.; \"Circuits, Matrices, and Linear Vector Spaces\"; pp. 57-63\n");
         printf ("Weinberg, L.; Chapter 4 in Shea, R. F., ed.; \"Amplifier Handbook\" pp. 4-27 - 4-31.\n");
         printf ("Lin, Pen-Min; \"Symbolic Network Analysis\"; pp. 115-117, pp. 213-219\n");
         printf ("Mayeda, W.; \"Graph Theory\"; ch. 7 and 8; pp. 252-341\n");
         printf ("\n");
         printf (" currently configured resource limits exclusive of requirements\n");
         printf ("for s and F parameters:\n");
         printf ("branch name list %d\n", emax-(s_active+s_passive));
         printf ("node list %d\n", nmax-s_nodes);
         printf ("passive edges %d\n", pmax-s_passive);
         printf ("active edges %d\n", actmax-s_active);
         printf ("\n");
         printf ("{sample sallen-key 3rd order lowpass:}\n");
         printf ("{optional intermediate results}\n");
         printf ("t 64\n");
         printf ("r1 1 2\n");
         printf ("r2 2 3\n");
         printf ("r3 3 4\n");
         printf ("c 2 5\n");
         printf ("c 3 6\n");
         printf ("c 4 5\n");
         printf ("{vcvs here}\n");
         printf ("u 6 5 4 5\n");
         printf ("f 6 5 1 5\n");
         printf ("end\n");
         return (1);
      } else if ((ch0 == 't') || (ch0 == 'T')) {
         scanf (" %u", &test);
         printf (" %u\n", test);
      } else {
         printf ("unrecognized element prefix-- '%s': must be r,l,c,g,y,z,gm,gy,o,u,b,rm,m,n,f,e\n", &branch[branches + 1][0]);
         printf ("h as input prints help\n");
         return (1);
      };



      if (p >= (pmax - s_passive)) {
         printf ("number of one-port elements(= %d) exceeds limit of %d\n or improper input applied\n", p, pmax - (1 + s_passive));
         return (1);
      };

      if (a >= (actmax - s_active)) {
         printf ("number of two-port elements(= %d) exceeds limit of %d\n or improper input applied\n", a, actmax - (1 + s_active));
         return (1);
      };

      if (n >= (nmax - s_nodes)) {
         printf ("number of nodes(= %d) exceeds limit of %d\n or improper input applied\n", a, nmax - (1 + s_nodes));
         return (1);
      };


   } printf ("\n\n");

   if (f == 0) {
      printf ("please provide a port node indication: f (lower case or upper case),\n");
      printf ("whitespace, +port1 node#, whitespace, -port1 node#, whitespace, +port2 node#, whitespace, -port2 node#\n\r");
      printf ("example:\n");
      printf ("f 1 0 2 4\n");

      return (1);
   };
   if ((test & PRINT_STATS) != 0) {
      printf (" currently configured resource limits:\n");
      printf ("branch name list limit:%d, used:%d\n", emax - (s_passive + s_active), branches);
      printf ("node list limit:%d, used:%d\n", nmax - s_nodes, n);
      printf ("passive edges limit:%d, used:%d\n", pmax - s_passive, p);
      printf ("active edges limit:%d, used:%d\n", actmax - s_active, a);
   };

   e = p;
   r = n - 1;

   eval_maps();



   if ((test & INPUT_DUMP) != 0) {
      printf ("Internal representation of data including generated elements and reassigned node#s:\n");
      printf ("edge,name,node#'s...\n");
      for (i = 1, j = 1; i <= p; i++, j++) {
         printf ("%3d %s %d %d\n", j, &branch[p0[i]][0], p1[i], p2[i]);
      };
      for (i = 1; i <= a; i++, j++) {
         printf ("%3d %s %d %d %d %d\n", j, &branch[a0[i]][0], a1[i], a2[i], a3[i], a4[i]);
      };
      printf ("%3d f %d %d %d %d\n", j, ni, nj, nk, nl);
      printf ("internal node#  input node#:\n");
      for (i = 1; i <= n; i++)
         printf ("         %5d    %d\n", i, nodelist[i]);
      printf ("\nedge namelist: index, value\n");
      for (i = 1; i <= branches; i++)
         printf ("%3d, %s\n", i, &branch[i][0]);
      printf ("\n");
   };

   return(0);
}

void analyze_network()
{
   int i, funs_evaluated;
   int port_node[4];

/* numerator and denominator indicators
   #define V     0
   #define wijkl 1
   #define wijij 2
   #define wklkl 3
   #define wklij 4
   #define uijkl 5
 */

/*
   transfer ratios:
   Vkl/Vij = wijkl/wijij
   Ikl/Iij = -wijkl/wklkl
   open circuit admittance functions:
   Iij/Vij = V/wijij
   Ikl/Vij = V/wklij
   Iij/Vkl = V/wijkl
   Ikl/Vkl = V/wklkl
   short circuit admittance functions:
   Iij/Vij = wklkl/uijkl
   Iij/Vkl = -wklij/uijkl
   Ikl/Vij = -wijkl/uijkl
   Ikl/Vkl = wijij/uijkl
 */

   port_node[0] = ni1;
   port_node[1] = nj1;
   port_node[2] = nk1;
   port_node[3] = nl1;


   for(i=0;i<6;i++)
      topological_fun[i] = 0;

   funs_evaluated = 0;

   for(i=0;i<(sizeof(function_table)/sizeof(f_tab));i++){


      evfn = function_table[i].fn_no;

      if(par_list[i] == 0) continue;

      if((evfn < 12) || (evfn > 19)){

        if(topological_fun[function_table[i].num_form] == 0){
          funs_evaluated++;
          topological_fun[function_table[i].num_form] = 1;
          ev_fn (function_table[i].num_form);
        };
        if(topological_fun[function_table[i].denom_form] == 0){
          funs_evaluated++;
          topological_fun[function_table[i].denom_form] = 1;
          ev_fn (function_table[i].denom_form);
        };

      };

   };
   
   if(funs_evaluated > 0) {

   for(i=0;i<(sizeof(function_table)/sizeof(f_tab));i++){

      evfn = function_table[i].fn_no;

      if(par_list[i] == 0) continue;

      if( ((evfn < 12) || (evfn >  19)) &&
         (topological_fun[function_table[i].num_form] != 0) &&
         (topological_fun[function_table[i].denom_form] != 0)){
         printf (" Function:\n %s( %d, %d; %d, %d) = %s( %d, %d)/%s( %d, %d) for %s( %d, %d)=0\n",
           function_table[i].fn_name,
           port_node[function_table[i].fn_p1p] ,port_node[function_table[i].fn_p1m],
           port_node[function_table[i].fn_p2p] ,port_node[function_table[i].fn_p2m] ,
           function_table[i].var1_name,
           port_node[function_table[i].var1_p1p] ,port_node[function_table[i].var1_p1m] ,
           function_table[i].var2_name,
           port_node[function_table[i].var2_p1p] ,port_node[function_table[i].var2_p1m] ,
           function_table[i].var3_name,
           port_node[function_table[i].var3_p1p] ,port_node[function_table[i].var3_p1m] );


         print_treebuf (function_table[i].sign, function_table[i].num_form, function_table[i].denom_form);
      };
   };


    release_topo_data();

  };




   for(i=0;i<6;i++)
      topological_fun[i] = 0;

   funs_evaluated = 0;

   for(i=0;i<(sizeof(function_table)/sizeof(f_tab));i++){

      if(par_list[i] == 0) continue;

      evfn = function_table[i].fn_no;

      if((evfn >= 12) && (evfn <= 19)){

        if(funs_evaluated == 0) embed_network_for_s_parameters();

        if(topological_fun[function_table[i].num_form] == 0){
          funs_evaluated++;
          topological_fun[function_table[i].num_form] = 1;
          ev_fn (function_table[i].num_form);
        };
        if(topological_fun[function_table[i].denom_form] == 0){
          funs_evaluated++;
          topological_fun[function_table[i].denom_form] = 1;
          ev_fn (function_table[i].denom_form);
        };

      };

   };

   if(funs_evaluated > 0) {

   for(i=0;i<(sizeof(function_table)/sizeof(f_tab));i++){

      evfn = function_table[i].fn_no;

      if(par_list[i] == 0) continue;

      if( ((evfn >= 12) && (evfn <=  19)) &&
         (topological_fun[function_table[i].num_form] != 0) &&
         (topological_fun[function_table[i].denom_form] != 0)){
         printf (" Function:\n %s( %d, %d; %d, %d) for %s normalization\n",
           function_table[i].fn_name,
           port_node[function_table[i].fn_p1p] ,port_node[function_table[i].fn_p1m],
           port_node[function_table[i].fn_p2p] ,port_node[function_table[i].fn_p2m] ,
           (normalization == 0)?"Y0":(normalization == 1)?"Z0":"unity");
         print_treebuf (function_table[i].sign, function_table[i].num_form, function_table[i].denom_form);
      };
   };

    release_topo_data();


  };


}
int 
nodemap (int node)
{
   int i;

   if (node > 999) {
      printf ("node label %d must be less than 1000\n", node);
      return (1);
   };

   for (i = 1; i <= n; i++) {
      if (nodelist[i] == node)
         return (i);
   };
   if (n >= (nmax-s_nodes)) {
      printf ("number of nodes(= %d) exceeds limit of %d\n or improper input applied\n", n, nmax - 1 - s_nodes);
      return (1);
   };
   n++;
   nodelist[n] = node;
   return (n);
}
int 
internal_node (void)
{
   if (n >= nmax) {
      printf ("number of nodes(= %d) exceeds limit of %d\n or improper input applied\n", n, nmax - 1);
      return (1);
   };

   n++;
   int_node++;
   nodelist[n] = int_node;
   return (n);
}


void embed_network_for_s_parameters(void)
{
/* -1<=>nk,-2<=>nl,-3<=>nj,-4<=>ni */
#define nodesubs(nnn) ((nnn == -1)?(nksave):((nnn == -2)?(nl):((nnn == -3)?(nj):((nnn == -4)?(nisave):(new_nodes[nnn])))))

   int nisave, nksave;
   int i, j, k;
   int new_nodes[10];

   int opamp_list[][4] = {
      {2, -3, 1, 2},
      {3, -3, -4, 4},
      {6, -2, 5, 6},
      {7, -2, -1, 8},
      {0, 0, 0, 0},
   };

/* vcis unused in embedding circuit */
   int vcis_list[][4] = {
      {0, 0, 0, 0},
   };

   int r_list[][2] = {
      {1, 3},
      {4, 3},
      {4, -3},
      {5, 7},
      {7, 8},
      {8, -2},
      {0, 0},
   };
   int ref_list[][2] = {
      {2, -4},
      {6, -1},
      {0, 0},
   };

printf("entered s-param\n");

   nisave = ni;
   nksave = nk;

   n--; /* back out dummy node for dummy edge */

   for(i=1;i<=8;i++) new_nodes[i] = internal_node();

   ni = nodesubs(1);
   nk = nodesubs(5);

   for(i=0;opamp_list[i][0] != 0;i++){

      opampcnt++;
      branches++;
      a++;
      a0[a] = branches;
      a1[a] = nodesubs(opamp_list[i][0]);
      a2[a] = nodesubs(opamp_list[i][1]);
      a3[a] = nodesubs(opamp_list[i][2]);
      a4[a] = nodesubs(opamp_list[i][3]);
      branch[branches][0] = 'o';
      branch[branches][1] = 0;


   };


   for(i=0;vcis_list[i][0] != 0;i++){


      branches++;
      a++;
      a0[a] = branches;
      a1[a] = nodesubs(vcis_list[i][0]);
      a2[a] = nodesubs(vcis_list[i][1]);
      a3[a] = nodesubs(vcis_list[i][2]);
      a4[a] = nodesubs(vcis_list[i][3]);
      branch[branches][0] = 'a';
      branch[branches][1] = 0;


   };

   for(i=0;r_list[i][0] != 0;i++){

      branches++;
      p++;
      p0[p] = branches;
      p1[p] = nodesubs(r_list[i][0]);
      p2[p] = nodesubs(r_list[i][1]);
      branch[branches][0] = 'a';
      branch[branches][1] = 0;


   };

   for(i=0;ref_list[i][0] != 0;i++){

      branches++;
      p++;
      p0[p] = branches;
      p1[p] = nodesubs(ref_list[i][0]);
      p2[p] = nodesubs(ref_list[i][1]);
      switch(normalization){
      case 0:
         branch[branches][0] = 'Y';
         branch[branches][1] = '0';
         branch[branches][2] = 0;
         break;
      case 1:
         branch[branches][0] = 'Z';
         branch[branches][1] = '0';
         branch[branches][2] = 0;
         break;
      default:
         branch[branches][0] = 'a';
         branch[branches][1] = 0;
      };



   };



/* dummy edge (handle) and extra node */
   p0[1] = 1;
   p1[1] = n + 1; /* update dummy node for this dummy edge */
   p2[1] = 1;
   branch[1][0] = 'x';
   branch[1][1] = 0;
   n++;

   e = p;
   r = n - 1;

   eval_maps();


   if ((test & INPUT_DUMP) != 0) {
      printf ("Internal representation of data including generated elements and reassigned node#s:\n");
      printf ("this includes embedding circuitry for evaluating scattering parameters.\n");
      printf ("edge,name,node#'s...\n");
      for (i = 1, j = 1; i <= p; i++, j++) {
         printf ("%3d %s %d %d\n", j, &branch[p0[i]][0], p1[i], p2[i]);
      };
      for (i = 1; i <= a; i++, j++) {
         printf ("%3d %s %d %d %d %d\n", j, &branch[a0[i]][0], a1[i], a2[i], a3[i], a4[i]);
      };
      printf ("%3d f %d %d %d %d\n", j, ni, nj, nk, nl);
      printf ("internal node#  input node#:\n");
      for (i = 1; i <= n; i++)
         printf ("         %5d    %d\n", i, nodelist[i]);
      printf ("\nedge namelist: index, value\n");
      for (i = 1; i <= branches; i++)
         printf ("%3d, %s\n", i, &branch[i][0]);
      printf ("\n");
   };


}

void 
evaluate_fn (int e, int a, int r, int topofun)
{
   int i, j, k, ch0, rc, passive_edges, minty_nodes, actual_nodes, active_edges,
     tree_form;

   evfn++;



      printf("Topological formula: (%s, %1d-trees) \n",
        topo_names[topofun],(topofun ==0)?1:((topofun == 5)?3:2));


   minty_nodes = r + 1;
   actual_nodes = r;
   passive_edges = e;
   dtag = 0;

   for (i = 1; i <= e; i++) {
      d0[i] = p0[i];
      dv1[i] = p1[i];
      dv2[i] = p2[i];
   };

   for (i = 1; i <= a; i++) {
      d0[e + i] = a0[i];
      dv1[e + i] = a1[i];
      dv2[e + i] = a2[i];
   };



      init(VARLIST(topofun));

   switch (topofun) {
   case V:
      active_edges = a;
      tree_form = ONETREES;
      for (i = 1; i <= a; i++) {
         j = dv1[e + i];
         dv1[e + i] = dv2[e + i];
         dv2[e + i] = j;
      };
      break;

   case wijkl:
      active_edges = a + 1;
      tree_form = TWOTREES;
      dv1[e + a + 1] = ni;
      dv2[e + a + 1] = nj;
      branch[branches + 1][0] = 'f';
      branch[branches + 1][1] = 0;
      d0[e + a + 1] = branches + 1;
      fmap[e + a + 1] = F1;
      break;

   case wijij:
      active_edges = a + 1;
      tree_form = TWOTREES;
      dv1[e + a + 1] = ni;
      dv2[e + a + 1] = nj;
      d0[e + a + 1] = branches + 1;
      branch[branches + 1][0] = 'f';
      branch[branches + 1][1] = 0;
      fmap[e + a + 1] = F1;
      break;

   case wklkl:
      active_edges = a + 1;
      tree_form = TWOTREES;
      dv1[e + a + 1] = nk;
      dv2[e + a + 1] = nl;
      d0[e + a + 1] = branches + 1;
      branch[branches + 1][0] = 'f';
      branch[branches + 1][1] = 0;
      fmap[e + a + 1] = F1;
      break;

   case wklij:
      active_edges = a + 1;
      tree_form = TWOTREES;
      dv1[e + a + 1] = nk;
      dv2[e + a + 1] = nl;
      branch[branches + 1][0] = 'f';
      branch[branches + 1][1] = 0;
      d0[e + a + 1] = branches + 1;
      fmap[e + a + 1] = F1;
      break;

   case uijkl:
      active_edges = a + 2;
      tree_form = THREETREES;
      dv1[e + a + 1] = ni;
      dv2[e + a + 1] = nj;
      d0[e + a + 1] = branches + 1;
      branch[branches + 1][0] = 'f';
      branch[branches + 1][1] = '1';
      branch[branches + 1][2] = 0;
      fmap[e + a + 1] = F1;
      dv1[e + a + 2] = nk;
      dv2[e + a + 2] = nl;
      d0[e + a + 2] = branches + 2;
      branch[branches + 2][0] = 'f';
      branch[branches + 2][1] = '2';
      branch[branches + 2][2] = 0;
      fmap[e + a + 2] = F2;
      break;

   };

   for (i = 1; i <= (passive_edges + active_edges); i++) {
      di1[i] = dv1[i];
      di2[i] = dv2[i];
   };




   switch (topofun) {
   case V:
      break;

   case wijkl:
      di1[e + a + 1] =  nk;
      di2[e + a + 1] =  nl;
      break;

   case wijij:
      break;

   case wklkl:
      break;

   case wklij:
      di1[e + a + 1] = ni;
      di2[e + a + 1] = nj;
      break;

   case uijkl:
      break;

   };

   for (i = 1; i <= a; i++) {
      di1[e + i] = a3[i];
      di2[e + i] = a4[i];
   };

/* strength reduction to reduce work in busiest loops */
   for (i = 1; i <= (passive_edges + active_edges); i++){
        dmap[i] = fmap[d0[i]];
        dpowermap[i] = powermap[d0[i]];
        dbranchmap[i] = branchmap[d0[i]];
        };



   if ((rc = product (minty_nodes, passive_edges + active_edges, actual_nodes,
     tree_form, e + a, topofun)) != 0) {
      printf ("insufficient memory to buffer common trees; remainder of function evaluation skipped\n");
      return;
   };


   if ((test & (PRINT_STATS)) != 0)
      printf ("opampcnt= %d\n", opampcnt);



/*   print_treebuf ((e + a));*/


}
int 
print_treebuf (int sign, int num_fun, int denom_fun)
{


      printf("Topological formula: %c(%s, %1d-trees) / (%s, %1d-trees)\n",
        (sign > 0)?' ':'-',
        topo_names[num_fun],(num_fun ==0)?1:((num_fun == 5)?3:2),
        topo_names[denom_fun],(denom_fun ==0)?1:((denom_fun == 5)?3:2));


     if(0 != copy(sign, numerator, num_fun) )
       return(1);
     if(0 != copy( 1, denominator, denom_fun) )
       return(1);




   if ((test & (DEBUG_512)) != 0) {
      release (NUM_VARS);
      release (DENOM_VARS);
      printf ("\n End:\n\n");
      return(0);
   };

   combineinit();

   if(0 != fixup () ){
      printf("insufficient memory to complete processing\n");
      printf ("\n Numerator:\n");
      printf ("\n Denominator:\n");
      printf ("\n End:\n\n");
      release (NUM_VARS);
      release (DENOM_VARS);
      return(1);
   };

   printf ("\n Numerator:\n");

   numcntxt = 0;
   printexp (NUM_VARS);



   printf ("\n Denominator:\n");

   numcntxt = 1;
   printexp (DENOM_VARS);

   release (NUM_VARS);
   release (DENOM_VARS);


   printf ("\n End:\n\n");

   return(0);

}
void 
treecheck (int r, int *d1, int *d2, int *tree, int *perm)
{
static   int stack[nmax], tag[nmax];
static   int depth, i, j, done, root_node, princip_node;

   for (i = 1; i <= r; i++) {
      tag[i] = 0;
   };

   perm[0] = 0;
   perm[1] = 0;
   stack[0] = 1;
   depth = 0;
   for (done = 0; done == 0;) {
      root_node = stack[depth];
      for (i = 1; i <= r; i++) {
         if (tag[i] == 1)
            continue;
         if (d1[tree[i]] == root_node) {
            tag[i] = 1;
            princip_node = d2[tree[i]];
            perm[princip_node] = tree[i];
            break;
         };
         if (d2[tree[i]] == root_node) {
            tag[i] = 1;
            princip_node = d1[tree[i]];
            perm[princip_node] = (-tree[i]);
            break;
         };
      };
      if (i <= r) {
         depth++;
         stack[depth] = princip_node;
      } else {
         if (depth > 0)
            depth--;
         else
            done = 1;
      };
   };
   return;
}

/*this code is from tree39h*/
int
product (int minty_nodes, int edges, int actual_nodes, int form, int br, int topofun)
{
static   int i, ii, j, k, m, nn, ncom, s, rc, sign, tmp, itreepermbuf[nmax], vtreepermbuf[nmax];
static   int tree[nmax];
static   int op, ch0;
static   int r;

static int opampcheck[actmax];

#define maxrunlength 10000
   int run, runstats[maxrunlength];

   for (run = 0; run < maxrunlength; run++)
      runstats[run] = 0;

for(i=0;i<actmax;i++)opampcheck[i] = 0;

printf("product:: minty_nodes:%d, edges:%d, actual_nodes:%d, form:%d, topofun:%d\n",
        minty_nodes, edges, actual_nodes, form, topofun);
   ncom = 0;

   istate = 0;
   vstate = 0;

   mn = minty_nodes;
   me = edges;
   mr = actual_nodes;
   mq = form;
   r = minty_nodes - 1;
   if ((test & (RAW_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS | DEBUG_128)) != 0) {
      printf("only %s-trees are accepted\n",(mq == TWOTREES)?"2":((mq == ONETREES)?"1":"3"));
      printf ("i-graph to generate i-trees:\nname:edge#,nameindex,node1,#node2#\n");
      for (i = 1; i <= me; i++)
         printf ("%5s: d0[%3d]=%3d, d1=%3d, d2=%3d\n", &branch[d0[i]][0], i, d0[i], di1[i], di2[i]);
      printf ("\n");
      printf ("\n");
      printf ("v-graph to check i-trees:\nname:edge#,nameindex,node1,#node2#\n")
;
      for (i = 1; i <= me; i++)
         printf ("%5s: d0[%3d]=%3d, d1=%3d, d2=%3d\n", &branch[d0[i]][0], i, d0[i], dv1[i], dv2[i]);
      printf ("\n");

};
/* search for common trees only */

      iminty ();
      {
         m = 0;
         run = 0;
         for (; (istate == 1) ;iminty ()) {

         ncom++;

         treecheck (r - 1, di1, di2, itree, &itreepermbuf[0]);
         treecheck (r - 1, dv1, dv2, itree, &vtreepermbuf[0]);

         if ((test & (TREE_PERM)) != 0) {
            printf (" common tree edges:");
            for (i = 1; i < r; i++)
               printf ("%3d(%s) ", itree[i],&branch[d0[itree[i]]][0]);
            printf ("\n");

            printf ("v-tree permutation: ");
            for (i = 2; i <= r; i++) {
               printf ("%3d(%s) ", vtreepermbuf[i],&branch[d0[ABS(vtreepermbuf[i])]][0]);
            };
            printf ("\n");
            printf ("i-tree permutation: ");
            for (i = 2; i <= r; i++) {
               printf ("%3d(%s) ", itreepermbuf[i],&branch[d0[ABS(itreepermbuf[i])]][0]);
            };
            printf ("\n");
         }
/* capture the principal node component of the sign computation */
         sign = 1;
         for (i = 2; i <= r; i++) {
            if (vtreepermbuf[i] < 0) {
               sign = sign * (-1);
               vtreepermbuf[i] = (-vtreepermbuf[i]);
            };
            if (itreepermbuf[i] < 0) {
               sign = sign * (-1);
               itreepermbuf[i] = (-itreepermbuf[i]);
            };
         };
/* capture the permutation component of the sign computation */
         for (i = 2; i <= r; i++) {
            {
               if (vtreepermbuf[i] != itreepermbuf[i]) {
                  sign = sign * (-1);
                  for (j = i + 1; j <= r; j++) {
                     if (vtreepermbuf[j] == itreepermbuf[i]) {
                        tmp = vtreepermbuf[j];
                        vtreepermbuf[j] = vtreepermbuf[i];
                        vtreepermbuf[i] = tmp;
                        break;
                     };
                  };
               };
            };
         };
         if ((test & (TREE_PERM)) != 0)
            printf ("resulting sign= %2d\n", sign);


      s = 0;
      for (i = 1, ii = 0, op = 0; i < r; i++) {
         s = s + dpowermap[itree[i]];

            if (dmap[itree[i]] != 0)
            continue;

         tree[ii] = itree[i];
         ii++;
      };


   if ((test & (RAW_TREES)) != 0) {
      printf (" accepted tree edges:");
      for (i = 0; i < ii; i++)
            printf ("%3d,%3d(%s) ", tree[i],d0[tree[i]],&branch[d0[tree[i]]][0]);
      printf("\n");
   };

      {

            if ((rc = getexpr (sign, s, ii, tree, br, VARLIST(topofun))) != 0) {
               release (VARLIST(topofun));
               topological_fun[topofun] = 0;
               printf ("insufficient memory for combining admittance terms\n");
               printf("raw i-trees  evaluated:%9d\n",i_nt);
               printf("qualified common trees:%9d\n",ncom);
               return (rc);
            };


       };

      };
   };

            if ((test & (PRINT_STATS)) != 0) {
               printf("raw i-trees  evaluated:%9d\n",i_nt);
               printf("qualified common trees:%9d\n",ncom);

            };
   return (0);
}


void 
iminty (void)
{				/* avoid parameter passing if possible */
static    int vd1, vd2, i, j, k, l, m, f0, f1, f2, op;
static    int v[nmax], edge[emax];

static   int stack[nmax], tag[nmax];
static   int depth, done, root_node, princip_node;

   switch (istate) {
   case 0:
      goto l10;
   case 1:
      goto l25;
   default:
      istate = 2;
      return;
   };
l10:
   istate = 1;
   ioprej = 0;

#if 0
   if ((test & (RAW_TREES | TREE_PERM | MIXED_UNITS | ADMITT_UNITS | DEBUG_128)) != 0) {
      printf("only %s-trees are accepted\n",(mq == TWOTREES)?"2":((mq == ONETREES)?"1":"3"));
      printf ("i-graph to generate i-trees:\nname:edge#,nameindex,node1,#node2#\n");
      for (i = 1; i <= me; i++)
         printf ("%5s: d0[%3d]=%3d, d1=%3d, d2=%3d\n", &branch[d0[i]][0], i, d0[i], di1[i], di2[i]);
      printf ("\n");
   };
#endif


   for (k = 1; k <= me; k++)
      edge[k] = 1;
   for (k = 1; k <= mn; k++)
      v[k] = 0;
   iit = 0;
   i = 1;
   i_nt = 0;
l15:
   for (k = 1; (k <= me); k++)
      if (edge[k] == 1)
         goto l40;
l25:
   if (iit <= 1) {
      istate = 2;
      return;
   };
   (iit)--;
   for (k = 1; k <= me; k++)
      edge[k] = iestk[iit][k];
   for (k = 1, i = 0; k <= mn; k++) {
      v[k] = ivstk[iit][k];
      if (v[k] > i)
         i = v[k];
   };
   i++;
   goto l15;

l40:
   l = k;
   edge[l] = 0;
   m = 0;
   for (k = 1; k <= me; k++)
      if (edge[k] != 0)
         m++;
   if (m >= (mn - 1)) {

      for (k = 1; k <= me; k++)
         iestk[iit][k] = edge[k];
      for (k = 1; k <= mn; k++)
         ivstk[iit][k] = v[k];
      (iit)++;
   }
l60:
   edge[l] = -1;
   vd1 = di1[l];
   vd2 = di2[l];

   if (v[vd1] == 0)
      goto l95;
   if (v[vd2] == 0)
      goto l105;

   l = v[vd1];
   m = v[vd2];

   if (l == m)
      goto l25;
   if (l > m) {
      k = m;
      m = l;
      l = k;
   };
   for (k = 1; k <= mn; k++) {
      if (v[k] > m)
         v[k]--;
      else if (v[k] == m)
         v[k] = l;
   };
   i--;
   goto l110;
l95:
   if (v[vd2] == 0)
      goto l100;
   v[vd1] = v[vd2];
   goto l110;
l100:
   v[vd1] = i;
   v[vd2] = i;
   i++;
   goto l15;
l105:
   v[vd2] = v[vd1];
l110:
   for (k = 1; k <= mn; k++)
      if (v[k] != 1)
         goto l15;
   (i_nt)++;
   l = 1;

   if ((test & (RAW_TREES)) != 0) {
      printf (" raw i-tree no. %5d edges:", i_nt);
      for (k = 2; k <= me; k++)
         if(edge[k] == -1) 
            printf ("%3d(%s) ", k,&branch[d0[k]][0]);
   };


/* q == 0 rejects all trees without 'f' branch; q == 1 rejects all trees with 'f' branch;
q == 2 rejects all trees without both 'f1' and 'f2 branches */

   f0 = 0;
   f1 = 0;
   f2 = 0;
   for (k = 2,op = 0; k <= me; k++)	/* dummy edge 1 is always present, so skip it */
      if (edge[k] == -1) {
         itree[l] = k;
         l++;
         f0 |= dmap[k];
         op += (dmap[k] & 0x1);
#if 0
         switch (dmap[k]) {
         case 0:
            break;
         case 1:
            f1 = 1;
            break;
         case 2:
            f2 = 1;
            break;
         case 3:
/*            if (mq == 1){
               if ((test & (RAW_TREES)) != 0)
                  printf(" rejected: not a 1-tree\n");
               goto l25;
            };
*/
            f0 = 1;
            break;
         case 4:
            break;
         case 5:
            op++;
            break;
         default:;
         };
#endif

      };

   if(opampcnt > op)
      {
      ioprej++;
      if ((test & (RAW_TREES)) != 0)
         printf(" rejected: opamp sorting\n");
      goto l25;
   };

   if ((f0 & TREEMASK) != mq) {
      if ((test & (RAW_TREES)) != 0) {
      printf("rejected: not a  %s-tree\n",(mq == TWOTREES)?"2":((mq == ONETREES)?"1":"3"));
      };
      goto l25;

   };

#if 0
   if (((mq == 0) || (mq == 1)) && (mq == f0))
   {
      if ((test & (RAW_TREES)) != 0) {
         if(mq == 1)printf(" rejected: not a 1-tree\n");
         else printf(" rejected: not a 2-tree\n");
      };
      goto l25;
   };

   if ((mq == 2) && ((f1 == 0) || (f2 == 0)))
   {
      if ((test & (RAW_TREES)) != 0)
         printf(" rejected: not a 3-tree\n");
      goto l25;

   };
#endif


/*this code is from tree39h*/
{
   for (i = 1; i < l; i++) {
      tag[i] = 0;
   };

   stack[0] = dv1[itree[1]];
   v[stack[0]] = 0;
   depth = 0;
   for (done = 0; done == 0;) {
      root_node = stack[depth];
      for (i = 1; i < l; i++) {
         if (tag[i] == 1)
            continue;
         if (dv1[itree[i]] == root_node) {
            tag[i] = 1;
            princip_node = dv2[itree[i]];
            break;
         };
         if (dv2[itree[i]] == root_node) {
            tag[i] = 1;
            princip_node = dv1[itree[i]];
            break;
         };
      };
      if (i < l) {
         depth++;
         stack[depth] = princip_node;
         v[princip_node] = 0;
      } else {
         if (depth > 0)
            depth--;
         else
            done = 1;
      };
   };
};

   for (k = 1; k < mn; k++)
      if (v[k] != 0){
         if ((test & (RAW_TREES)) != 0) printf(" rejected: no corresponding v-tree\n");
         goto l25;
       }

   if ((test & (RAW_TREES)) != 0) printf(" accepted\n");

   return;

}

/* code to combine like terms */ void 
combineinit (void)
{
   int i;


   for (i = 0; i < maxsym; i++)
      cancelfactor[i] = 0x7fffffff;

   cancelcoeff = 0;
   cancelsign = -1;
   cancelexp = 0x7fffffff;


}
void 
init (PARAMETERLIST)
{
   int i, j, k;
   int (*poly)[maxdegree], (*termlist)[maxsym + 2];

   *maxtermlist = maxterm * maxdegree;
   *maxpoly = maxterm;
   *ptermlist = (void *) malloc ((*maxtermlist) * sizeof (int) * (maxsym + 2));
   *ppoly = (void *) malloc ((*maxpoly) * sizeof (int) * maxdegree);

   poly = *ppoly;
   termlist = *ptermlist;

   for (j = 0; j < maxdegree; j++)
      for (k = 0; k < (*maxpoly); k++)
         poly[k][j] = 0;

   for (i = 0; i < (*maxtermlist); i++)
      for (j = 0; j <= (maxsym + 1); j++)
         termlist[i][j] = 0;
   for (j = 0; j < maxdegree; j++)
      degterms[j] = 0;
   *toterms = 0;

}
int copy(int sign, int dest, int source)
{
   int i, j, k;
   int (*polys)[maxdegree], (*termlists)[maxsym + 2];
   int (*polyd)[maxdegree], (*termlistd)[maxsym + 2];

vars[dest].termlist = (void *) malloc ((vars[source].maxterms) * sizeof (int) * (maxsym + 2));

if (0 == vars[dest].termlist){
printf("can't allocate space for termlist\n");
   return(1);
   };

vars[dest].poly = (void *) malloc ((vars[source].maxpoly) * sizeof (int) * maxdegree);

if (0 == vars[dest].poly)
 {
  free(vars[dest].termlist);
  printf("can't allocate space for poly\n");
  return(1);
  };


   polyd = vars[dest].poly;
   termlistd = vars[dest].termlist;
   polys = vars[source].poly;
   termlists = vars[source].termlist;

   for (j = 0; j < maxdegree; j++)
      for (k = 0; k < (vars[source].maxpoly); k++)
         polyd[k][j] = polys[k][j];

   vars[dest].maxpoly = vars[source].maxpoly;

   for (i = 0; i < (vars[source].maxterms); i++)
      for (j = 0; j <= (maxsym + 1); j++)
         termlistd[i][j] = termlists[i][j];

   if (sign < 0)
     for (i = 0; i < (vars[source].maxterms); i++)
       termlistd[i][maxsym] = (-termlistd[i][maxsym]);

   vars[dest].maxterms = vars[source].maxterms;

   for (j = 0; j < maxdegree; j++)
      vars[dest].degterms[j] = vars[source].degterms[j];

   vars[dest].toterms = vars[source].toterms;


return(0);

}

int resize(void ** ptr, int * count, int module)
{
   int tmpsize;
   void * tmpptr;
   if ((test & (PRINT_STATS)) != 0)printf ("orig size =%d, orig ptr =0x%08x\n",(*count) ,(*ptr) );
   tmpsize = 2 * (*count);
   tmpptr = (void *) realloc ((*ptr), tmpsize * module);

   if ((tmpptr != (void *) 0)) {
      if ((test & (PRINT_STATS)) != 0)printf ("num of elements =%d, pointer =0x%08x\n", tmpsize, tmpptr);
      *ptr = tmpptr;
      *count = tmpsize;
      return(0);
   } else {
      tmpsize = (tmpsize + (*count)) / 2;
      while (tmpsize > (*count)) {
         tmpptr = (void *) realloc ((*ptr), tmpsize * module);

         if ((tmpptr != (void *) 0)) {
            if ((test & (PRINT_STATS)) != 0)printf ("num of elements =%d, pointer =0x%08x\n", tmpsize, tmpptr);
            *ptr = tmpptr;
            *count = tmpsize;
            return(0);
         } else {
            tmpsize = (tmpsize + (*count)) / 2;
            if (tmpsize == (*count)){
               if ((test & (PRINT_STATS)) != 0)printf("resize failed for insufficient memory\n");
               return (1);
            };
         };
      }
   };
   printf("resize failed for insufficient memory\n");
   return(1);

}

void release_topo_data(void)
{
      int i;
      for(i=V;i<=uijkl;i++)
        if (topological_fun[i] !=0) release(VARLIST(i));

}


void
release (PARAMETERLIST)
{
   free (*ptermlist);
   free (*ppoly);

}
void 
eval_branchtag (PARAMETERLIST)
{
/* branchtag collects all the impedance-unit factors to be divided out later */

   int i, j, k, l, m, n;
   int (*poly)[maxdegree], (*termlist)[maxsym + 2];


   poly = *ppoly;
   termlist = *ptermlist;


   if ((test & (DEBUG_512)) != 0)
      return;


   for (i = 0, n = 0; i < maxdegree; i++) {
      if (degterms[i] > 0) {
         for (j = 0; (j < degterms[i]); j++) {
            l = poly[j][i];
            if ((termlist[l][maxsym] != 0)) {
                  for (k = 0; k < impedancecount; k++) {
                     m = impedancelist[k];
                     if (termlist[l][m] > branchtag[m])
                        branchtag[m] = termlist[l][m];
                  };
            };

         };
      };
   };



   return;
}

void eval_maps(void)
{
int i,j, ch0;

   for (j = 1; j <= branches; j++) {
      branchmap[j] = getsym (j);
   };

   for (i = 0; i < maxsym; i++)
      branchtag[i] = 0;

   impedancecount = 0;


   for (i = 1; i <= branches; i++) {

      j = branchmap[i] /*getsym (i, (e + a))*/;
      ch0 = branch[j][0];

      if ((ch0 == 'c') ||
        (ch0 == 'C') ||
        (ch0 == 'l') ||
        (ch0 == 'L') ||
        (ch0 == 'm') ||
        (ch0 == 'M'))
         powermap[i] = 1;
      else
         powermap[i] = 0;

      if (ch0 == 'a')
         fmap[i] = UNITY;
      else
      {
         if ((ch0 == 'o') ||
           (ch0 == 'O'))
            fmap[i] = OP;
         else
            fmap[i] = 0;
      };

      if (i > j)
         continue;
      if ((ch0 == 'r') ||
        (ch0 == 'R') ||
        (ch0 == 'z') ||
        (ch0 == 'Z') ||
        (ch0 == 'l') ||
        (ch0 == 'L') ||
        (ch0 == 'm') ||
        (ch0 == 'M')) {
         impedancelist[impedancecount] = j;
         impedancecount++;
      };
   };

}



int
getexpr (int sign, int deg, int factors, int *factorlist, int edges, PARAMETERLIST)
{
   int i, j, k, l, m, n;
   int u,v;
   int (*poly)[maxdegree], (*termlist)[maxsym + 2];
   int tmpsize,origsize;
   void *tmpptr;

   poly = *ppoly;
   termlist = *ptermlist;

   for (i = 0; i < maxsym; i++)
      term[i] = 0;

   for (j = 0; j < factors; j++) {
      i = dbranchmap[factorlist[j]];
      term[i] = term[i] + 1;
   };
/*debug int denfn,numfn,inptno,evfn; */
   if ((test & (DEBUG_256)) != 0) {
      printf ("%2d,%2d,%2d,%2d,%2d,%2d:", inptno, evfn, numfn, denfn, deg, sign);
      for (j = 0; j < factors; j++)
         printf (",%3d", factorlist[j]);
      printf ("\n");
   };
   if ((test & (DEBUG_512)) != 0)
      return (0);

   if (degterms[deg] == 0) {
      poly[0][deg] = *toterms;
      for (i = 0; i < maxsym; i++)
         termlist[*toterms][i] = term[i];
      termlist[*toterms][maxsym] = sign;
      termlist[*toterms][maxsym + 1] = 0;
      if (((*toterms)+1) >= (*maxtermlist)) {
         {
            if ((test & (PRINT_STATS)) != 0)printf("resizing termlist ");
            tmpsize = (*maxtermlist);
            origsize = tmpsize;
            tmpptr = (*ptermlist);
            if(0 != resize((void *)&tmpptr, &tmpsize, sizeof (int) * (maxsym + 2)) ){
               return(1);
            };
            *ptermlist = tmpptr;
            *maxtermlist = tmpsize;
            termlist = *ptermlist;
         } for (i = origsize; i < (*maxtermlist); i++)
            for (j = 0; j <= (maxsym + 1); j++)
               termlist[i][j] = 0;

      }
      (*toterms)++;
      degterms[deg] = 1;

   } else {

      for (m = (degterms[deg]-1)/2, u = 0, v = (degterms[deg]-1),n = 0; (u <= v) && (n >= 0); m = (u+v)/2) {


         k = poly[m][deg];
         for (i = 0; i < maxsym; i++){
            if (termlist[k][i] < term[i]){
               v = m-1;
               n = m;
               break;
            }
            else
               if (termlist[k][i] > term[i]){
               u = m+1;
               n = m+1;
               break;
            }
            else
               n = -1;
         };
         if (n < 0) {
            termlist[k][maxsym] += sign;

         };
      };
      if (n >= 0) {
         k = degterms[deg];
         if(n < k)
            for(i=k; i>n; i--)
               poly[i][deg] = poly[i-1][deg];

         poly[n][deg] = *toterms;
         for (i = 0; i < maxsym; i++)
            termlist[*toterms][i] = term[i];
         termlist[*toterms][maxsym] = sign;
         termlist[*toterms][maxsym + 1] = 0;

         if (((*toterms)+1) >= (*maxtermlist)) {
            {

               if ((test & (PRINT_STATS)) != 0)printf("resizing termlist ");
               tmpsize = (*maxtermlist);
               origsize = tmpsize;
               tmpptr = (*ptermlist);
               if(0 != resize((void *)&tmpptr, &tmpsize, sizeof (int) * (maxsym + 2)) ){
                  return(1);
               };
               *ptermlist = tmpptr;
               *maxtermlist = tmpsize;
               termlist = *ptermlist;

            } for (i = origsize; i < (*maxtermlist); i++)
               for (j = 0; j <= (maxsym + 1); j++)
                  termlist[i][j] = 0;

         }
         (*toterms)++;
         degterms[deg]++;
         if ((degterms[deg]+1) >= (*maxpoly)) {
            {

               if ((test & (PRINT_STATS)) != 0)printf("resizing poly ");
               tmpsize = (*maxpoly);
               origsize = tmpsize;
               tmpptr = (*ppoly);

               if(0 != resize((void *)&tmpptr, &tmpsize, sizeof (int) * maxdegree) ){
                  return(1);
               };
               *ppoly = tmpptr;
               *maxpoly = tmpsize;
               poly = *ppoly;
            } for (j = 0; j < maxdegree; j++)
               for (k = origsize; k < (*maxpoly); k++)
                  poly[k][j] = 0;


         };


      };
   };

   return (0);
}
void 
printexp (PARAMETERLIST)
{
   int i, j, k, l, m, n;

   int (*poly)[maxdegree], (*termlist)[maxsym + 2];

   poly = *ppoly;
   termlist = *ptermlist;





   if (cancelcoeff == 0)
      cancelcoeff = 1;

   if ((test & (ADMITT_UNITS)) != 0) {
      printf ("common factors to cancel in all terms of both num and denom:\n");
      printf ("%s%d ", cancelsign > 0 ? "+" : "-", cancelcoeff);
      for (k = 0; k < maxsym; k++)
         if ((cancelfactor[k]) > 0) {
            if ((cancelfactor[k]) > 1)
               printf (" %s^%-2d", &branch[k][0], (cancelfactor[k]));
            else
               printf (" %s   ", &branch[k][0]);
         };

      printf ("\n");
   };



   for (i = 0, n = 0; i < maxdegree; i++) {
      if (degterms[i] > 0) {
         for (j = 0, k = 0; (j < degterms[i]) && (k == 0); j++) {
            l = poly[j][i];
            if ((termlist[l][maxsym] != 0))
               k = 1;

         };

         if (k == 1) {
            n++;
            printf ("\n + s^%-2d (\n", i - cancelexp);
            for (j = 0; j < degterms[i]; j++) {
               l = poly[j][i];
               if ((termlist[l][maxsym] == 0))
                  continue;
               if ((test & (DEBUG_1024)) != 0)
                  printf ("%2d,%2d,%2d,%2d,%2d:", inptno, evfn, numcntxt, i - cancelexp, (((cancelsign * termlist[l][maxsym] / cancelcoeff) < 0) ? -1 : 1));
               else
                  printf ("         ");
               {
                  if (((cancelsign * termlist[l][maxsym] / cancelcoeff) > 1) ||
                    ((cancelsign * termlist[l][maxsym] / cancelcoeff) < 0))
                     printf ("%3d", cancelsign * termlist[l][maxsym] / cancelcoeff);
                  else {
                     if ((test & (DEBUG_1024)) != 0)
                        printf ("  1");
                     else
                        printf ("   ");
                  };
               };
               for (k = 0, m = 0; k < maxsym; k++) {
                  if ((termlist[l][k] - cancelfactor[k]) > 0) {
                     m = 1;
                     if ((termlist[l][k] - cancelfactor[k]) > 1)
                        printf (" %s^%-2d", &branch[k][0], (termlist[l][k] - cancelfactor[k]));
                     else
                        printf (" %s   ", &branch[k][0]);
                  };
               };
               if ((m == 0) && ((cancelsign * termlist[l][maxsym] / cancelcoeff) == 1))
                  printf (" 1");
               printf ("\n");
            };
            printf ("        )");
         };
      };
   };
   if (n == 0)
      printf ("        0");
   printf ("\n");
}
int 
fixup (void)
{
   int j, k;


   eval_branchtag (NUM_VARS);
   eval_branchtag (DENOM_VARS);

   for (j = 1; j <= (e + a); j++) {
      if (branchtag[j] == 0)
         continue;
      else {
         k = branchmap[j] /*getsym (j, (e + a))*/;
         if (j > k) {
            branchtag[j] = 0;
            branchtag[k]++;
         };
      };
   };

   if ((test & (MIXED_UNITS)) != 0) {
      printf ("these ohm-unit factors modify all terms of both num and denom:\n");
      for (k = 1; k <= (e + a); k++)
         if ((branchtag[k]) > 0) {
            if ((cancelfactor[k]) > 1)
               printf (" %s^%-2d", &branch[k][0], (branchtag[k]));
            else
               printf (" %s   ", &branch[k][0]);
         };

      printf ("\n");
   for (j = 1; j <= (e + a); j++) branchtag[j] = 0;
   printf("note that s*L, s*M, and R represent s/L, s/M, and 1/R in the following function:\n");
   };

   if(0 != correctunits (NUM_VARS) )
      return(1);
   if(0 != correctunits (DENOM_VARS) )
      return(1);
   makecancellationlist (NUM_VARS);
   makecancellationlist (DENOM_VARS);

   return(0);

}
int 
correctunits (PARAMETERLIST)
{
   int i, ii, j, k, l, m, n, s;
   int kk;
   int u,v;
   int (*poly)[maxdegree], (*termlist)[maxsym + 2];
   int tmpsize, origsize;
   void *tmpptr;


   poly = *ppoly;
   termlist = *ptermlist;



   for (i = 0; i < maxdegree; i++) {
      if (degterms[i] > 0) {
         for (j = 0, k = 0; (j < degterms[i]) && (k == 0); j++) {
            l = poly[j][i];
            if (		/*(termlist[l][maxsym] != 0)&& */
              (termlist[l][maxsym + 1] == 0))
               k = 1;

         };

         if (k == 1) {
            for (j = 0; j < degterms[i]; j++) {
               l = poly[j][i];
               if (	/*(termlist[l][maxsym] == 0) || */
                 (termlist[l][maxsym + 1] == 1))
                  continue;
               termlist[l][maxsym + 1] = 1;



               for (k = 0; k < maxsym; k++) {
                  if ((branchtag[k] != 0)) {
                     termlist[l][k] = branchtag[k] - termlist[l][k];
                  };
               };
               for (k = 0, s = 0; k < maxsym; k++) {
                  if (powermap[k] != 0)
                     s += termlist[l][k];
               };
               if (s != i) {
                  if (degterms[s] == 0) {
                     poly[0][s] = l;
                     degterms[s] = 1;

                  } else {
                     for (m = 0, n = 0; (m < degterms[s]) && (n == 0); m++) {
                        k = poly[m][s];
                        if (termlist[k][maxsym + 1] == 0)
                           continue;
                        for (ii = 0; ii < maxsym; ii++)
                           if (termlist[k][ii] != termlist[l][ii])
                              break;
                        if (ii == maxsym) {
                           n = 1;
                           termlist[k][maxsym] += termlist[l][maxsym];

                        };
                     };
                     if (n == 0) {
                        k = degterms[s];
                        if ((degterms[s]+1) >= (*maxpoly)) {
                           {





                              if ((test & (PRINT_STATS)) != 0)printf("resizing poly ");
                              tmpsize = (*maxpoly);
                              origsize = tmpsize;
                              tmpptr = (*ppoly);

                              if(0 != resize((void *)&tmpptr, &tmpsize, sizeof (int) * maxdegree) ){
                                 return(1);
                              };
                              *ppoly = tmpptr;
                              *maxpoly = tmpsize;
                              poly = *ppoly;
                           } for (ii = 0; ii < maxdegree; ii++)
                              for (n = origsize; n < (*maxpoly); n++)
                                 poly[n][ii] = 0;

                        };
                        poly[k][s] = l;
                        degterms[s]++;


                     };
                  };
                  if ((j + 1) < degterms[i])
                     for (ii = (j + 1); ii < degterms[i]; ii++)
                        poly[ii - 1][i] = poly[ii][i];
                  degterms[i]--;
                  j--;
               };
            };
         };
      };
   };




   for (i = 0; i < maxdegree; i++) {
      if (degterms[i] > 0) {
         for (j = (degterms[i] - 1); (j >= 0); j--) {
            l = poly[j][i];
            if (		/*(termlist[l][maxsym] != 0)&& */
              (termlist[l][maxsym + 1] == 1)) {

               for(v=0,u=(j-1),k=((j-1)/2);(u>=v);k=((u+v)/2)){
                  m = poly[k][i];

                  for (ii = 0; ii < maxsym; ii++){
                     if (termlist[l][ii] < termlist[m][ii])
                     {
                        v = k + 1;
                        break;
                     }
                     else
                        if (termlist[l][ii] > termlist[m][ii])
                     {
                        u = k - 1;
                        break;
                     };

                  };

                  if (ii == maxsym) {

                     termlist[m][maxsym] += termlist[l][maxsym];
                     if ((j + 1) < degterms[i])
                        for (ii = (j + 1); ii < degterms[i]; ii++)
                           poly[ii - 1][i] = poly[ii][i];
                     degterms[i]--;

                     break;
                  };




               };


            };

         };
      };
   };

   return(0);


}
void 
makecancellationlist (PARAMETERLIST)
{
   int i, j, k, l, m, n;

   int (*poly)[maxdegree], (*termlist)[maxsym + 2];


   poly = *ppoly;
   termlist = *ptermlist;

   for (i = 0; i < maxdegree; i++) {
      if (degterms[i] > 0) {
         for (j = 0, k = 0; (j < degterms[i]) && (k == 0); j++) {
            l = poly[j][i];
            if ((termlist[l][maxsym] != 0))
               k = 1;

         };

         if (k == 1) {
            if (cancelexp > i)
               cancelexp = i;
            for (j = 0; j < degterms[i]; j++) {
               l = poly[j][i];
               if ((termlist[l][maxsym] == 0))
                  continue;

               for (k = 0; k < maxsym; k++) {
                  if ((termlist[l][k] < cancelfactor[k]))
                     cancelfactor[k] = termlist[l][k];
               };

               if ((termlist[l][maxsym] > 0))
                  cancelsign = 1;

               if (((n = termlist[l][maxsym]) < 0))
                  n = -n;
               if ((cancelcoeff > 0) && (cancelcoeff != n)) {
                  if (cancelcoeff > n) {
                     m = n;
                     n = cancelcoeff;
                  } else
                     m = cancelcoeff;

                  for (; n > m;) {
                     for (; n > m;)
                        n -= m;
                     l = m;
                     m = n;
                     n = l;
                  };
                  cancelcoeff = n;
               } else
                  cancelcoeff = n;
            };
         };
      };
   };
}
int 
getsym (int i)
{

   int j, k, l, m;


   k = 1;
   while (k <= branches) {
      m = 0;
      while ((branch[k][m] == branch[i][m]) && (branch[i][m] != 0))
         m++;
      if (branch[k][m] == branch[i][m]) {
         return k;
      };
      k++;
   };
   printf ("can't find %s <=> branch[%d] in %d branches\n", &branch[i][0], i, branches);
   return i;

}
